/* $Id: head.c,v 4.4.4.1 1992/02/23 21:25:39 sob PATCH_4 sob $
 *
 * $Log: head.c,v $
 * Revision 4.4.4.1  1992/02/23  21:25:39  sob
 * Patch level 4
 *
 * Revision 4.4.3.1  1992/02/01  03:09:32  sob
 * Release 4.4 Patchlevel 3
 *
 * Revision 4.4  1991/09/09  20:18:23  sob
 * release 4.4
 *
 *
 * 
 */
/* This software is Copyright 1991 by Stan Barber. 
 *
 * Permission is hereby granted to copy, reproduce, redistribute or otherwise
 * use this software as long as: there is no monetary profit gained
 * specifically from the use or reproduction of this software, it is not
 * sold, rented, traded or otherwise marketed, and this copyright notice is
 * included prominently in any copy made. 
 *
 * The author make no claims as to the fitness or correctness of this software
 * for any use whatsoever, and it is provided as is. Any use of this software
 * is at the user's own risk. 
 */

/* #define USENOV		/* move to config.h later */
#include "EXTERN.h"
#include "common.h"
#include "artio.h"
#include "bits.h"
#ifdef SERVER
#include "server.h"
#endif
#include "util.h"
#ifdef HCACHE
#include "term.h"	/* for hcache miss spinner */
#ifdef USENOV
#include "nov.h"
#include "rn.h"		/* ngname */
#endif
#endif /* HCACHE */
#include "INTERN.h"
#include "head.h"

bool first_one;		/* is this the 1st occurance of this header line? */

static char htypeix[26] =
    {0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0};

void
head_init()
{
    register int i;

    for (i=HEAD_FIRST+1; i<HEAD_LAST; i++)
	htypeix[*htype[i].ht_name - 'a'] = i;
}

#ifdef DEBUGGING
void
dumpheader(where)
char *where;
{
    register int i;

    printf("header: %d %s", parsed_art, where);

    for (i=0; i<HEAD_LAST; i++) {
	printf("%15s %4d %4d %03o\n",htype[i].ht_name,
	    htype[i].ht_minpos,
	    htype[i].ht_maxpos,
	    htype[i].ht_flags) FLUSH;
    }
}
#endif

int
set_line_type(bufptr,colon)
char *bufptr;
register char *colon;
{
    char lc[LONGKEY+3];
    register char *t, *f;
    register int i, len;

    if (colon-bufptr > LONGKEY+2)
	return SOME_LINE;

    for (t=lc,f=bufptr; f<colon; f++, t++) {
	if (isspace(*f))
	/* guard against space before : */
	    break;
	*t = isupper(*f) ? tolower(*f) : *f;
    }
    *t = '\0';
    f = lc;				/* get lc into register */
    len = t - f;

    /* now scan the headtype table, backwards so we don't have to supply an
     * extra terminating value, using first letter as index, and length as
     * optimization to avoid calling subroutine strEQ unnecessarily.  Hauls.
     */
    
    if (islower(*f)) {
	for (i = htypeix[*f - 'a']; *htype[i].ht_name == *f; --i) {
	    if (len == htype[i].ht_length && strEQ(f, htype[i].ht_name)) {
		return i;
	    }
	}
    }
    return SOME_LINE;
}

void
start_header(artnum)
ART_NUM artnum;
{
    register int i;

#ifdef DEBUGGING
    if (debug & 4)
	dumpheader("start_header\n");
#endif
    for (i=0; i<HEAD_LAST; i++) {
	htype[i].ht_minpos = -1;
	htype[i].ht_maxpos = 0;
    }
    in_header = SOME_LINE;
    first_one = FALSE;
#ifdef ASYNC_PARSE
    parsed_art = artnum;
#endif
}

bool
parseline(art_buf,newhide,oldhide)
char *art_buf;
int newhide, oldhide;
{
    if (*art_buf == ' ' || *art_buf == '\t') {
					/* header continuation line? */
#ifdef HCACHE
	if ((hc_artnum>0) && (hc_line_doing == in_header))
	    hc_line_continue(in_header,art_buf);
#endif
	return oldhide;
    }
    else {				/* maybe another header line */
	char *s;

#ifdef HCACHE
	hc_line_doing = 0;
#endif
	if (first_one) {		/* did we just pass 1st occurance? */
	    first_one = FALSE;
	    htype[in_header].ht_maxpos = artpos;
					/* remember where line left off */
	}
	s = index(art_buf,':');
	if (s == Nullch) {
			    /* is it the end of the header? */
	    htype[PAST_HEADER].ht_minpos =
		(*art_buf == '\n') ? ftell(artfp) : artpos;
			    /* remember where body starts */
	    in_header = PAST_HEADER;
	}
	else {	/* it is a new header line */
	    in_header = set_line_type(art_buf,s);
	    first_one = (htype[in_header].ht_minpos < 0);
	    if (first_one) {
#ifdef HCACHE
		if ((hc_artnum>=0) && hcache[in_header] && 
	 (HC_ELEM(in_header,AOFFSET(hc_artnum)) == Nullch)) {
		    hc_line_start(in_header,s);
		    hc_line_doing = in_header;
		}
#endif
		htype[in_header].ht_minpos = artpos;
	    }
#ifdef DEBUGGING
	    if (debug & 4)
		dumpheader(art_buf);
#endif
	    if (htype[in_header].ht_flags & HT_HIDE)
		return newhide;
	}
    }
    return FALSE;			/* don't hide this line */
}

#ifdef ASYNC_PARSE
int
parse_maybe(artnum)
ART_NUM artnum;
{
    char tmpbuf[LBUFLEN];

    if (parsed_art == artnum)
	return 0;
    if (artnum > lastart)
	return -1;
    /* no maybe about it now */
    if (artnum<absfirst || artnum>lastart)
	return -1;	/* not worth trying */
#ifdef AVAILBITS
    if (is_unavailable(artnum))
	return -1;	/* don't even try... */
#endif
#ifdef SERVER
    if (nntpopen(artnum,GET_HEADER) == Nullfp) {
#else
    if (artopen(artnum) == Nullfp) {
#endif
#ifdef AVAILBITS
	/* let them know its unavailable */
	unavail_set(artnum);
#endif
	return -1;
    }
#ifdef AVAILBITS
    avail_set(artnum);
#endif
    start_header(artnum);
#ifdef HCACHE
    hc_artnum = artnum;
#endif
    while (in_header) {
	artpos = ftell(artfp);
	if (fgets(tmpbuf,LBUFLEN,artfp) == Nullch)
	    break;
	parseline(tmpbuf,FALSE,FALSE);
    }
    in_header = PAST_HEADER;
#ifdef HCACHE
    {
	int i;

	/* set all empty ones to the empty header */
	for (i=0;i<HEAD_LAST;i++)
	    if (hcache[i] && (HC_ELEM(i,AOFFSET(hc_artnum)) == Nullch))
		HC_ELEM(i,AOFFSET(hc_artnum)) = hc_empty_entry;
    }
    hc_artnum = -1;
    hc_line_doing = 0;
#endif
    return 0;
}
#endif

/* get the subject line for an article */
char *
fetchsubj(artnum,current_subject,copy)
ART_NUM artnum;				/* article to get subject from */
bool_int current_subject;		/* is it in a parsed header? */
bool_int copy;				/* do you want it savestr()ed? */
{
    static char fetchsubj_buf[LBUFLEN];	/* reconsider? */
    char *s = Nullch, *t;
#ifdef SERVER
    static int xhdr = 1;		/* Can we use xhdr command? */
    int eoo;				/* End of server output */
#endif /* SERVER */

#ifdef CACHESUBJ
    if (!subj_list) {
	register ART_NUM i;
#ifndef lint
	subj_list =
	  (char**)safemalloc((MEM_SIZE)((OFFSET(lastart)+2)*sizeof(char *)));
#endif /* lint */
	for (i=0; i<=OFFSET(lastart); i++)
	    subj_list[i] = Nullch;
    }
#endif /* CACHESUBJ */

#ifdef HCACHE
    if (hcache && hcache[SUBJ_LINE]) {
	fetchsubj_buf[0] = '\0';
	s = fetchlines(artnum,SUBJ_LINE);
#ifdef CACHESUBJ
	if (!subj_list[OFFSET(artnum)])
	    if (s)
		subj_list[OFFSET(artnum)] = savestr(s);
	    else /* !s (use an empty string) */
		subj_list[OFFSET(artnum)] = savestr(fetchsubj_buf);
#endif /* CACHESUBJ */
	if (copy) {
	    return(s);	/* their responsibility */
	}
	if (s) {
	    strcpy(fetchsubj_buf,s);
	    free(s);
	}
	return(fetchsubj_buf);
    }
#endif /* HCACHE */

#ifdef CACHESUBJ
    if (artnum < absfirst || artnum > lastart)
	s = nullstr;
    else
	s = subj_list[OFFSET(artnum)];
#endif
    if (s == Nullch) {
	if (current_subject) {
	    s = fetchlines(artnum,SUBJ_LINE);
#ifdef CACHESUBJ
	    subj_list[OFFSET(artnum)] = s;
#endif
	}
	else {
	    s = safemalloc((MEM_SIZE)LBUFLEN);
	    *s = '\0';
#ifdef SERVER
	    if (xhdr) {
	    	sprintf(ser_line, "XHDR subject %ld", artnum);
	    	put_server(ser_line);
#ifdef DEBUGGING
		if (debug & DEB_NNTP)
		    printf(">%s\n", ser_line) FLUSH;
#endif
		if (nntp_get(ser_line, sizeof (ser_line)) >= 0) {
#ifdef DEBUGGING
			if (debug & DEB_NNTP)
			    printf("<%s\n", ser_line) FLUSH;
#endif
			if (ser_line[0] == CHAR_FATAL) {
				fprintf(stderr,"\nrrn: %s\n",ser_line);
				finalize(1);
/*			    xhdr = 0; */
			} else {
			    while (nntp_get(ser_line, sizeof (ser_line)) >= 0) {
#ifdef DEBUGGING
				if (debug & DEB_NNTP)
				    printf("<%s\n", ser_line) FLUSH;
#endif
				if (ser_line[0] == '.')
				    break;
				else {
				    t = index(ser_line, ' ');
				    if (t++) {
					strcpy(s, t);
					if (t = index(s, '\r'))
						*t = '\0';
				    }
				}
			    }
			}
		} else {
		    fprintf(stderr,
			"\nrrn: Unexpected close of server socket.\n");
		    finalize(1);
		}
	    }

	    if (!xhdr) {
		sprintf(ser_line, "HEAD %ld", artnum);
		put_server(ser_line);
#ifdef DEBUGGING
		if (debug & DEB_NNTP)
		    printf(">%s\n", ser_line) FLUSH;
#endif
		eoo = 0;
		if (nntp_get(ser_line, sizeof (ser_line)) >= 0 && 
			ser_line[0] == CHAR_OK) {
		    do {
			if (nntp_get(s, LBUFLEN) < 0 || (*s == '.')) {
			strcpy(s, "Title: \n");
			eoo = 1;
		        }
		    } while (strnNE(s,"Title:",6) && strnNE(s,"Subject:",8));

		    if (!eoo)
			while (nntp_get(ser_line, sizeof (ser_line)) >= 0 &&
				ser_line[0] != '.');
		    t = index(s,':')+1;
		    while (*t == ' ') t++;
		    strcpy(s, t);
	        }
	    }
#else /* not SERVER */
	    if (artopen(artnum) != Nullfp) {
		do {
		    if (fgets(s,LBUFLEN,artfp) == Nullch)
			strcpy(s, "Title: \n");
		} while (strnNE(s,"Title:",6) && strnNE(s,"Subject:",8));

		s[strlen(s)-1] = '\0';
		t = index(s,':')+1;
		while (*t == ' ') t++;
		strcpy(s, t);
	    }
#endif
	    s = saferealloc(s, (MEM_SIZE)strlen(s)+1);
#ifdef CACHESUBJ
	    subj_list[OFFSET(artnum)] = s;
#endif 
	}
    }
#ifdef CACHESUBJ
    if (copy) {
	t = savestr(s);
	return t;
    }
    else
	return s;
#else
    if (copy)
	return s;
    else {
	safecpy(cmd_buf,s,CBUFLEN);	/* hope this is okay--we're */
	free(s);
	return cmd_buf;			/* really scraping for space here */
    }
#endif
}

/* get header lines from an article */

char *
fetchlines(artnum,which_line)
ART_NUM artnum;				/* article to get line from */
int which_line;				/* type of line desired */
{
    char *newbuf, *t, tmp_buf[LBUFLEN];
    register ART_POS curpos;
    int size;
    register ART_POS firstpos;
    register ART_POS lastpos;
    
#ifdef HCACHE
    if (t = hc_fetchline(artnum,which_line))
	return(savestr(t));
#endif

#ifdef ASYNC_PARSE
    if (parse_maybe(artnum))
	artnum = 0;
#ifdef HCACHE
    else
	/* maybe the parse_maybe brought it into the cache */
	if (t = hc_fetchline(artnum,which_line))
	    return(savestr(t));
#endif
#endif
    firstpos = htype[which_line].ht_minpos;
    lastpos = htype[which_line].ht_maxpos;
    if (!artnum || firstpos < 0) {
	newbuf = safemalloc((unsigned int)1);
	*newbuf = '\0';
	return newbuf;
    }
#ifdef SERVER
    if (!artnum || firstpos < 0 || nntpopen(artnum,GET_HEADER) == Nullfp) {
#else
    if (!artnum || firstpos < 0 || artopen(artnum) == Nullfp) {
#endif
	newbuf = safemalloc((unsigned int)1);
	*newbuf = '\0';
	return newbuf;
    }
#ifndef lint
    size = lastpos - firstpos + 1;
#else
    size = Null(int);
#endif /* lint */
#ifdef DEBUGGING
    if (debug && (size < 1 || size > 1000)) {
	printf("Firstpos = %ld, lastpos = %ld\n",(long)firstpos,(long)lastpos);
	gets(tmp_buf);
    }
#endif
    newbuf = safemalloc((unsigned int)size);
    *newbuf = '\0';
    fseek(artfp,firstpos,0);
    for (curpos = firstpos; curpos < lastpos; curpos = ftell(artfp)) {
	if (fgets(tmp_buf,LBUFLEN,artfp) == Nullch)
	    break;
	if (*tmp_buf == ' ' || *tmp_buf == '\t')
	    t = tmp_buf;
	else {
	    t = index(tmp_buf,':');
	    if (t == Nullch)
		break;
	    t++;
	}
	while (*t == ' ' || *t == '\t') t++;
	safecat(newbuf,t,size);
    }
    return newbuf;
}


/* returns FROM: line from either the cache or a fetch */
/* returns NULL or empty string if it was unavailable for whatever reason... */
/* #ifdef SERVER code mostly copied from fetchsubj */
char *
get_from_line(a)
ART_NUM a;
{
    static char get_from_buf[LBUFLEN];
    int i;		/* used in #ifdef'ed code */
    char *s1;
#ifdef SERVER
    char *s,*t;
    int xhdr=1;	/* assume xhdr--if you don't have xhdr then set xhdr=0 */
#endif
#ifdef HCACHE
    if (hcache && hcache[FROM_LINE]) {
	get_from_buf[0] = '\0';
	if (s1 = fetchlines(a,FROM_LINE)) {
	    strcpy(get_from_buf,s1);
	    free(s1);
	}
	return(get_from_buf);
    }
#endif
#ifdef CACHEFROM
    if (!from_list) {		/* no cache before now */
	from_list = (char**)safemalloc((AOFFSET(lastart)+2)*sizeof(char*));
	for (i=0 ; i<=AOFFSET(lastart) ; i++)
	    from_list[i] = Null(char *);
    }
    if (!from_list[AOFFSET(a)]) {	/* no old entry */
#endif /* CACHEFROM */
#ifdef SERVER
/* UGH!  Need to reindent this sometime... */
	    if (xhdr) {
		s = get_from_buf;	/* temp space... */
		*s = '\0';
	    	sprintf(ser_line, "XHDR from %ld", a);
	    	put_server(ser_line);
#ifdef DEBUGGING
		if (debug & DEB_NNTP)
		    printf(">%s\n", ser_line) FLUSH;
#endif
		if (nntp_get(ser_line, sizeof (ser_line)) >= 0) {
#ifdef DEBUGGING
			if (debug & DEB_NNTP)
			    printf("<%s\n", ser_line) FLUSH;
#endif
			if (ser_line[0] == CHAR_FATAL) {
				fprintf(stderr,"\nrrn: %s\n",ser_line);
				finalize(1);
/*			    xhdr = 0; */
			} else {
			    while (nntp_get(ser_line, sizeof (ser_line)) >= 0) {
#ifdef DEBUGGING
				if (debug & DEB_NNTP)
				    printf("<%s\n", ser_line) FLUSH;
#endif
				if (ser_line[0] == '.')
				    break;
				else {
				    t = index(ser_line, ' ');
				    if (t++) {
					strcpy(s, t);
					if (t = index(s, '\r'))
						*t = '\0';
				    }
				}
			    }
			}
		} else {
		    fprintf(stderr,
			"\nrrn: Unexpected close of server socket.\n");
		    finalize(1);
		}
#ifdef CACHEFROM
	        from_list[AOFFSET(a)] = savestr(s);
#endif
		return(s);	/* points to space within get_from_buf */
	    }
	    else {	/* !xhdr */
#endif /* SERVER */
	/* fetchlines saves the string */
	s1 = fetchlines(a,FROM_LINE);
#ifdef CACHEFROM
	from_list[AOFFSET(a)] = s1;
	return(s1);
#else
	strcpy(get_from_buf,s1);
	free(s1);
	return(get_from_buf);
#endif
#ifdef SERVER
	    }		/* end if big if/else stmt above. reindent later */
#endif
#ifdef CACHEFROM
    } /* if (not cached) */
    return(from_list[AOFFSET(a)]);
#endif
}

#ifdef HCACHE
/* should all of this be put in hcache.[ch] files? */

/* last artnum that could be cached */
int hc_lastart;	/* "size" of hcache */

static char **hc_strings;	/* contains saved strings */
static int hc_str_used;		/* actually filled */
static int hc_str_alloc;	/* pointers allocated */

char *
hc_save(oldstring)
char *oldstring;
{
    if (hc_str_used == hc_str_alloc) {	/* need to grow */
	hc_str_alloc += HC_STR_CHUNK;
	if (hc_str_alloc == HC_STR_CHUNK) /* not allocated before */
	    hc_strings = (char**)safemalloc((HC_STR_CHUNK)*sizeof(char*));
	else	/* already something allocated */
	    hc_strings = (char**)saferealloc((char*)hc_strings,
				(hc_str_alloc)*sizeof(char*));
    }
    hc_strings[hc_str_used] = savestr(oldstring);
    return(hc_strings[hc_str_used++]);
}

void
hc_free()
{
    int i;
    for (i=0;i<hc_str_used;i++)
	free(hc_strings[i]);
    if (hc_str_alloc>0)
	free(hc_strings);
    hc_strings = (char**)NULL;
    hc_str_used = hc_str_alloc = 0;
}

void
hc_init()
{
    int i,j;
#ifdef USENOV
    register struct novart *novp;	/* overview db misc */
#endif

    /* clear the string tracking (should be deallocated) */
    hc_str_used = hc_str_alloc = 0;

    /* note: zeroth entry is wasted */
    /* CAA 11/18/92:
     * I used to have an error here (HEAD_LAST-1) instead of HEAD_LAST.
     * Thanks to Pete Bevin for a timely debugging dump. */
    hcache = (char***)safemalloc((HEAD_LAST)*sizeof(char**));
    for (i=0;i<HEAD_LAST;i++)
	hcache[i] = (char**)NULL;
    for (i=0;i<HEAD_LAST;i++) {
	if ((i==SUBJ_LINE) || (i==FROM_LINE) ||
/* #ifdef UNDEF /* */
	/* If you want to play with the multi-line article displays then
	 * include this code.  It won't waste too much memory, however,
	 * since most articles have neither summary nor keyword lines.
	 */
	    (i==SUMRY_LINE) || (i==KEYW_LINE) ||
	/* for early moderation testing */
/*	    (i==MESSID_LINE) || /* */
/* #endif /* */
	    (i==NGS_LINE)) {
	    hcache[i] = (char**)safemalloc((AOFFSET(lastart)+2)
					*sizeof(char*));
	    for (j=0;j<=AOFFSET(lastart);j++)
		HC_ELEM(i,j) = (char *)NULL;
	}
    }
    hc_lastart = lastart;
#ifdef USENOV
    if (!use_nov(ngname))
	return;		/* can't go further */
    /* if we got this far, then cache some more! */
    for (i=0;i<HEAD_LAST;i++) {
	if ((!hcache[i]) &&
	    ((i==MESSID_LINE) ||
	    (i==REFS_LINE))
	   ) {
	    hcache[i] = (char**)safemalloc((AOFFSET(lastart)+2)
					*sizeof(char*));
	    for (j=0;j<=AOFFSET(lastart);j++)
		HC_ELEM(i,j) = (char *)NULL;
	}
    }
    for (i=0, novp = novall(novgp); novp != NULL; novp = novp->a_nxtnum) {
	if ((!novp->a_num) || (j=atoi(novp->a_num))==0)
	    continue;		/* no article */
	i++;
	if (novp->a_subj)
	    HC_ELEM(SUBJ_LINE,AOFFSET(j)) = novp->a_subj;
	if (novp->a_from)
	    HC_ELEM(FROM_LINE,AOFFSET(j)) = novp->a_from;
	if (novp->a_msgid)
	    HC_ELEM(MESSID_LINE,AOFFSET(j)) = novp->a_msgid;
	if (novp->a_refs)
	    HC_ELEM(REFS_LINE,AOFFSET(j)) = novp->a_refs;
    }
    printf("Nov database has %d articles.\n",i) FLUSH;
#endif /* USENOV */
}

void
hc_grow(oldlast,newlast)
ART_NUM oldlast,newlast;
{
    int i,j;

    if (!hcache)	/* not initialized yet */
	return;
    printf("\nhc_grow(%d,%d)\n",oldlast,newlast) FLUSH;
    for (i=0;i<HEAD_LAST;i++) {
	if (hcache[i]) {
	   hcache[i] = (char**)saferealloc((char *)hcache[i],
				(AOFFSET(newlast)+2)*sizeof(char*));
	    for (j=AOFFSET(oldlast)+1;j<=AOFFSET(newlast);j++)
		HC_ELEM(i,j) = (char *)NULL;
	}
    }
    hc_lastart = newlast;
    /* insert NOV growth here? (problematic?) */
    printf("\nLeaving: hc_grow(%d,%d)\n",oldlast,newlast) FLUSH;
}

bool spin_active INIT(FALSE);
char spinchars[] = {'|','/','-','\\'};
int spin_place;		/* represents place in spinchars array */
int spin_count;		/* counter for when to spin */
int spin_level INIT(0);	/* used to allow non-interfering nested spins */

void
hc_setspin(flag)
bool_int flag;
{
    if (flag)
	spin_level++;
    else
	spin_level--;
    if (spin_level>0)
	spin_active = TRUE;
    else {
	spin_active = FALSE;
	if (spin_place>1) {	/* we have spun at least once */
	/* get rid of spin character */
	    putchar(' ');
	    backspace();
	    fflush(stdout);
	}
    }
    if (flag && spin_level==1) {	/* first activation */
	spin_count = 1;	/* not 0 to prevent immediate spin display */
	spin_place = 1; /* start with slash... */
    }
}

void
hc_spin(count)
int count;		/* modulus for the spin... */
{
    if (!(spin_count++%count)) {
	putchar(spinchars[spin_place++%4]);
	backspace();
	fflush(stdout);
    }
}

/* returns the line from article a. (or Nullch) */
/* do not mess with what this returns... */
char *
hc_fetchline(a,line)
ART_NUM a;
int line;
{
    if (a<absfirst || a>hc_lastart)
	return(Nullch);
    if (!hcache)
	return(Nullch);
    if (hcache[line] && HC_ELEM(line,AOFFSET(a)))
	return(HC_ELEM(line,AOFFSET(a)));
    if (spin_active)
	hc_spin(25);	/* spin every 25 cache misses */
#ifdef SERVER
    if (
#ifdef ASYNC_PARSE
	(a != parsed_art) &&	/* don't mess with the current art. */
#endif
	(hcache[line])) {
	    hc_nntp_fastfill(a);
	    return(HC_ELEM(line,AOFFSET(a)));
    }
#endif /* SERVER */
    return(Nullch);
}

void
hc_line_start(line,s)
int line;		/* which line are we doing? */
char *s;		/* points at the colon... */
{
    char *s1,*s2;
    char tmpbuf[LBUFLEN];

#ifdef UNDEF
    printf("hc_line_start(%d,|%s|)(%d)\n",
	line,s,hc_artnum) FLUSH;
#endif
    s++;	/* skip over the colon */
    while ((*s == ' ') || (*s == '\t'))
	s++;
    s1 = (char*)strncpy(tmpbuf,s,LBUFLEN);
    s2 = s1+strlen(s1)-1;	/* points at the '\n' */
    if (*s2 == '\n')
	*s2 = '\0';
    HC_ELEM(line,AOFFSET(hc_artnum)) = hc_save(tmpbuf);
}

/* BUG!?!  I don't think it should *replace* the first one... */

void
hc_line_continue(line,s)
int line;		/* which line are we doing? */
char *s;		/* points at the start of the continuation line */
{
    char *s1,*s2;
    char tmpbuf[LBUFLEN];

#ifdef UNDEF
    printf("hc_line_continue(%d,|%s|)(%d)\n",
	line,s,hc_artnum) FLUSH;
#endif
    while ((*s == ' ') || (*s == '\t'))
	s++;
    s1 = (char*)strncpy(tmpbuf,s,LBUFLEN);
    s2 = s1+strlen(s1)-1;	/* points at the '\n' */
    if (*s2 == '\n')
	*s2 = '\0';
    HC_ELEM(line,AOFFSET(hc_artnum)) = hc_save(tmpbuf);
}

void
hc_clean()
{
    int i,j;
#ifdef UNDEF
    printf("HCACHE cleanup...");
    fflush(stdout);
#endif
    hc_free();		/* free allocated strings */
    if (hcache) {
	for (i=0;i<HEAD_LAST;i++)
	    if (hcache[i])
		free(hcache[i]);
	free(hcache);
    }
    hcache = NULL;
    hc_lastart = -1;
#ifdef UNDEF
    printf("Done.") ;
    fflush(stdout);
#endif
}

#ifdef SERVER
void
hc_nntp_fastfill(a)
ART_NUM a;
{
    char *line;		/* pointer to start of good line */
    int header_in = -1;	/* what header are we currently in... */
    int i;

    hc_artnum = a;	/* let hcache routines know... */

    sprintf(ser_line, "HEAD %ld", (long)a);
#ifdef DEBUGGING
    if (debug & DEB_NNTP)
	printf(">%s\n", ser_line) FLUSH;
#endif
    put_server(ser_line);		/* ask the server for the article */
    if (nntp_get(ser_line, sizeof(ser_line)) < 0) {
	fprintf(stderr, "\nrrn: Unexpected close of server socket.\n");
	finalize(1);
    }
#ifdef DEBUGGING
    if (debug & DEB_NNTP)
	printf("<%s\n", ser_line) FLUSH;
#endif
    if (*ser_line == CHAR_FATAL) {	/* Fatal error */
		fprintf(stderr,"\nrrn: %s\n",ser_line);
		finalize(1);
	}
    if (*ser_line != CHAR_OK) {		/* and get it's reaction */
#ifdef AVAILBITS
	/* let them know its unavailable */
	    unavail_set(a);
#endif
        return;
    }
#ifdef AVAILBITS
    /* let them know its available */
    avail_set(a);
#endif

    for (;;) {
        if (nntp_get(ser_line, sizeof(ser_line)) < 0) {
	    fprintf(stderr, "\nrrn: Unexpected close of server socket.\n");
	    finalize(1);
	}
	if (ser_line[0] == '.' && ser_line[1] == '\0')
		break;
	line = (ser_line[0] == '.' ? ser_line + 1 : ser_line);
	if ((line[0] == '\t') || (line[0] == ' ')) {	/* continuation */
	    if ((header_in>=0) && (hcache[header_in]))
		hc_line_continue(header_in,line);
	    continue;	/* read in another line */
	}
	{
	    char *s;
	    if (s = index(line,':')) {
		header_in = set_line_type(line,s);
		if (header_in == SOME_LINE) {
		    /* for now no extra line types */
		    header_in = -1;
		    continue;	/* read in next line */
		}
		if(hcache[header_in])	/* cached header type */
		    hc_line_start(header_in,s);
	    }
	}
    } /* for */
    {
	ART_NUM j = AOFFSET(a);

	for (i=0;i<HEAD_LAST;i++)
	    if (hcache[i] && (HC_ELEM(i,j) == (char *)NULL))
		HC_ELEM(i,j) = hc_empty_entry;
    }
}
#endif /* SERVER */
#endif /* HCACHE */
