/* This file Copyright 1992 by Clifford A. Adams */
/* sacmd.c
 *
 * main command loop
 */

#include "EXTERN.h"
#include "common.h"
#ifdef SCAN
/* absfirst, lastart, oneless() */
#include "bits.h"
#include "term.h"	/* for LINES */
#include "head.h"	/* for get_from_line */
#include "help.h"	/* online help */
#ifdef USETHREADS
#include "ngdata.h"		/* for ThreadedGroup */
#endif
#include "scan.h"
#include "smisc.h"
#include "spage.h"
#include "sdisp.h"
#include "scanart.h"
#include "samain.h"
#include "samisc.h"
#include "sadisp.h"
#include "sadesc.h"
#include "sathread.h"
#ifdef SCORE
#include "score.h"
#endif
#include "INTERN.h"
#include "sacmd.h"

/* several basic commands are already done by s_docmd (Scan level) */
/* interprets command in buf, returning 0 to continue looping,
 * a condition code (negative #s) or an art# to read.  Also responsible
 * for setting refresh flags if necessary.
 */
int
sa_docmd()
{
    ART_NUM a;		/* article pointed to */
    ART_NUM b;		/* misc. artnum */
    int i,j;		/* for misc. purposes */
    bool flag;		/* misc */

    a = (ART_NUM)page_ents[*s_ptr_page_line].entnum;
    switch(*buf) {
#ifdef USETHREADS
	case '+':	/* enter thread selector */
	    if (!ThreadedGroup) {
		s_beep();
		return(0);
	    }
	    buf[0] = '+';	/* fake up command for return */
	    buf[1] = '\0';
	    sa_art = a; /* give it somewhere to point */
	    return(SA_FAKE);	/* fake up the command. */
#endif
#ifdef SCORE
	case '!':	/* debug */
	    sc_dumpscores();
	    sa_keypause();
	    *s_ref_all = TRUE;
	    break;
#endif
#ifdef UNDEF
	case '@':	/* debug */
	    printf("\n*s_top_ent: %d\n",*s_top_ent);
	    printf("index,article,startline,lines,status\n");
	    for (i=0;i<=*s_bot_ent;i++) {
		printf("%2d %6d %2d %2d |%s|\n",i,page_ents[i].entnum,
			page_ents[i].start_line,
			page_ents[i].lines,
			sa_get_statchars(page_ents[i].entnum,1)) FLUSH;
	    }
	    sa_keypause();
	    *s_ref_all = TRUE;
	    break;
#endif
#ifdef UNDEF
	case '&':	/* specific debug */
	    printf("\n%d,%d,%d (absfirst,firstart,lastart)\n",
		   absfirst,firstart,lastart);
	    i = sa_first();
	    printf("sa_first() = %d, sa_last() = %d\n",
		   i,sa_last());
	    printf("sa_next(sa_first()) = %d\n",
		   sa_next(i));
	    i = sa_next(i);
	    for (j=sa_first();j<i;j++) {
		if (is_unavailable(j))
		    printf("%d is unavailable\n",j) FLUSH;
	    }
	    sa_keypause();
	    *s_ref_all = TRUE;
	    sa_debug_value = 1;
	    break;
#endif /* UNDEF */
#ifdef SCORE
	case 'K':	/* kill below a threshold */
	    *buf = ' ';				/* for finish_cmd() */
	    if (!sa_finish_cmd("Kill below or equal score:"))
		break;
	    /* make **sure** that there is a number here */
	    i = atoi(buf+1);
	    if (i == 0) {			/* it might not be a number */
		char *s;
		s = buf+1;
		if (!((*s=='0') || (((*s=='+')||(*s=='-'))&&(s[1]=='0')))) {
		    /* text was not a numeric 0 */
		    s_beep();
		    return(0);			/* beep but do nothing */
		}
	    }
	    sc_kill_threshold(i);
	    *s_refill = TRUE;
	    *s_ref_top = TRUE;	/* refresh # of articles */
	    break;
#endif /* SCORE */
	case 'D':	/* kill unmarked "on" page */
	    for (i=0;i<=*s_bot_ent;i++)
    /* This is a difficult decision, with no obviously good behavior. */
    /* Do not kill threads with the first article marked, as it is probably
     * not what the user wanted.
     */
		if (!sa_marked(page_ents[i].entnum) || !sa_mode_fold)
		   (void)sa_art_cmd(sa_mode_fold,SA_KILL_UNMARKED,
				    page_ents[i].entnum);
/* consider: should it start reading? */
	    b = sa_readmarked_elig();
	    if (b) {
		sa_clearmark(b);
		return(b);
	    }
	    *s_ref_top = TRUE;	/* refresh # of articles */
	    *s_refill = TRUE;
	    break;
	case 'J':	/* kill marked "on" page */
	    /* If in "fold" mode, kill threads with the first article
	       marked */
	    if (sa_mode_fold) {
		for (i=0;i<=*s_bot_ent;i++) {
		    if (sa_marked(page_ents[i].entnum))
			(void)sa_art_cmd(TRUE,SA_KILL,page_ents[i].entnum);
		}
	    } else
		for (i=0;i<=*s_bot_ent;i++)
		    (void)sa_art_cmd(FALSE,SA_KILL_MARKED,page_ents[i].entnum);
	    *s_refill = TRUE;
	    *s_ref_top = TRUE;	/* refresh # of articles */
	    break;
	case 'X':	/* kill unmarked (basic-eligible) in group */
	    i = sa_first();
	    if (!i)
		break;
	    if (!sa_basic_elig(i))
		i = sa_next(i);
#ifdef HCACHE
	    hc_setspin(TRUE);	/* may take awhile */
#endif
	    /* New action: if in fold mode, will not delete an article which
	       has a thread-prior marked article. */
	    for (;i;i=sa_next(i)) {
		if ((!sa_marked(i)) && (!was_read(i))) {
		    if (sa_mode_fold) {		/* new semantics */
			ART_NUM j;
			/* make j == 0 if no prior marked articles in
			   the thread. */
			for (j=i;j;j=sa_subj_thread_prev(j))
			    if (sa_marked(j))
				break;
			if (j)	/* there was a marked article */
			    continue;	/* article selection loop */
		    }
		    oneless(i);
		}
	    }
#ifdef HCACHE
	    hc_setspin(FALSE);
#endif
	    b = sa_readmarked_elig();
	    if (b) {
		sa_clearmark(b);
		return(b);
	    }
	    *s_refill = TRUE;
	    *s_ref_top = TRUE;	/* refresh # of articles */
	    break;
#ifdef SCORE
	case 'o':
	    if (sa_mode_order<sa_mode_order_max)
		sa_mode_order++;
	    else
		sa_mode_order = 1;
	    if ((sa_mode_order==2)	/* score order */
		&& !sc_initialized)
		    sa_mode_order = 1;	/* nope... (maybe allow later) */
	    /* if we go into score mode, make sure score is displayed */
	    if ((sa_mode_order==2) && !sa_mode_desc_score)
		sa_mode_desc_score = TRUE;
	    s_go_top_ents();
	    *s_refill = TRUE;
	    *s_ref_bot = TRUE;
	    break;
	case 'O':	/* change article sorting order */
	    if (sa_mode_order != 2) { /* not in score order */
		s_beep();
		break;
	    }
	    score_newfirst = !score_newfirst;
	    sc_sort();
	    s_go_top_ents();
	    *s_refill = TRUE;
	    break;
	case 'R':	/* rescore articles */
	    if (!sc_initialized)
		break;
	    /* clear to end of screen */
	    clear_rest();
	    *s_ref_all = TRUE;	/* refresh everything */
	    printf("\nRescoring articles...\n") FLUSH;
	    sc_rescore();
	    s_go_top_ents();
	    *s_refill = TRUE;
	    eat_typeahead();	/* stay in control. */
	    break;
	case Ctl('e'):		/* edit scorefile for group */
	    /* clear to end of screen */
	    clear_rest();
	    *s_ref_all = TRUE;	/* refresh everything */
	    sc_score_cmd("e");	/* edit scorefile */
	    eat_typeahead();	/* stay in control. */
#endif /* SCORE */
	case '\t':	/* TAB: toggle threadcount display */
	    sa_mode_desc_threadcount = !sa_mode_desc_threadcount;
	    *s_ref_desc = 0;
	    break;
	case 'a':	/* toggle author display */
	    sa_mode_desc_author = !sa_mode_desc_author;
	    *s_ref_desc = 0;
	    break;
#ifdef SCORE
	case 'A':	/* toggle author status type */
	    sa_auth_showstat = !sa_auth_showstat;
	    *s_ref_status = 0;
	    break;
#endif
	case '#':	/* toggle article # display */
	    sa_mode_desc_artnum = !sa_mode_desc_artnum;
	    *s_ref_desc = 0;
	    break;
	case 'U':	/* toggle unread/unread+read mode */
	    sa_mode_read_elig = !sa_mode_read_elig;
	    if (sa_eligible(sa_first()) || sa_next_elig(sa_first())) {
#ifdef SCORE
#ifdef PENDING
		if (sa_mode_read_elig) {
		    sc_fill_read = TRUE;
		    sc_fill_min = lastart;
		}
		if (!sa_mode_read_elig)
		    sc_fill_read = FALSE;
#endif
#endif
		*s_ref_top = TRUE;
		s_rub_ptr();
		s_go_top_ents();
		*s_refill = TRUE;
	    } else { /* quit out: no articles */
		return(SA_QUIT);
	    }
	    break;
	case 'F':	/* Fold */
	    sa_mode_fold = !sa_mode_fold;
	    *s_refill = TRUE;
	    *s_ref_top = TRUE;
	    break;
	case 'f':	/* follow */
	    sa_follow = !sa_follow;
	    *s_ref_top = TRUE;
	    break;
	case 'Z':	/* Zero (wipe) selections... */
	    for (i=0;i<=AOFFSET(lastart);i++)
		sa_per_art[i].sa_art_status =
		  (sa_per_art[i].sa_art_status & 0xfd);
	    *s_ref_status = 0;
	    if (!sa_mode_zoom)
		break;
	    *s_ref_all = TRUE;	/* otherwise won't be refreshed */
	    /* if in zoom mode, turn it off... */
	    /* FALL THROUGH */
	case 'z':	/* zoom mode toggle */
	    sa_mode_zoom = !sa_mode_zoom;
		/* toggle mode if there will remain an eligible article */
	    if (sa_eligible(sa_first()) || sa_next_elig(sa_first())) {
		*s_ref_top = TRUE;
		s_go_top_ents();
		*s_refill = TRUE;
	    } else {
		s_beep();
		sa_mode_zoom = !sa_mode_zoom;	/* toggle it right back */
	    }
	    break;
	case 'j':	/* junk just this article */
			/* vi mode no longer supported */
	    (void)sa_art_cmd(FALSE,SA_KILL,a);
	    /* later refill the page */
	    *s_refill = TRUE;
	    *s_ref_top = TRUE;	/* refresh # of articles */
	    break;
	case 'n':	/* next art */
	case ']':
	    s_rub_ptr();
	    if (*s_ptr_page_line<(*s_bot_ent)) { /* more on page... */
		*s_ptr_page_line +=1;
	    }
	    else {
		if (!sa_next_elig(page_ents[*s_bot_ent].entnum)) {
		    s_beep();
		    return(0);
		}
		s_go_next_page();	/* will jump to top too... */
	    }
	    break;
	case 'k':	/* kill subject thread */
			/* vi mode no longer supported */
	case ',':	/* might later clone TRN ',' */
	    (void)sa_art_cmd(TRUE,SA_KILL,a);
	    *s_refill = TRUE;
	    *s_ref_top = TRUE;	/* refresh # of articles */
	    break;
	case Ctl('n'):	/* follow subject forward */
	    i = subj_thread(a);
#ifdef HCACHE
	    hc_setspin(TRUE);
#endif
	    for (b=sa_next_elig(a);b;b=sa_next_elig(b))
		if (i==subj_thread(b))
		    break;
#ifdef HCACHE
	    hc_setspin(FALSE);
#endif
	    if (!b) {	/* no next w/ same subject */
		s_beep();
		return(0);
	    }
	    for (i = *s_ptr_page_line+1;i<=*s_bot_ent;i++)
		if (page_ents[i].entnum == b) {	/* art is on same page */
		    s_rub_ptr();
		    *s_ptr_page_line = i;
		    return(0);
		}
	    /* article is not on page... */
	    (void)s_fillpage_forward(b);  /* fill forward (*must* work) */
	    *s_ptr_page_line = 0;
	    break;
	case Ctl('a'):	/* next article with same author... */
	/* good for scoring */
	    b = sa_wrap_next_author(a);
	    /* rest of code copied from next-subject case above */
	    if (!b || (a == b)) {	/* no next w/ same subject */
		s_beep();
		return(0);
	    }
	    for (i=0;i<=*s_bot_ent;i++)
		if (page_ents[i].entnum == b) {	/* art is on same page */
		    s_rub_ptr();
		    *s_ptr_page_line = i;
		    return(0);
		}
	    /* article is not on page... */
	    (void)s_fillpage_forward(b);  /* fill forward (*must* work) */
	    *s_ptr_page_line = 0;
	    break;
	case Ctl('p'):	/* follow subject backwards */
	    i = subj_thread(a);
#ifdef HCACHE
	    hc_setspin(TRUE);
#endif
	    for (b=sa_prev_elig(a);b;b=sa_prev_elig(b))
		if (i==subj_thread(b))
		    break;
#ifdef HCACHE
	    hc_setspin(FALSE);
#endif
	    if (!b) {	/* no next w/ same subject */
		s_beep();
		return(0);
	    }
	    for (i = *s_ptr_page_line-1;i>=0;i--)
		if (page_ents[i].entnum == b) {	/* art is on same page */
		    s_rub_ptr();
		    *s_ptr_page_line = i;
		    return(0);
		}
	    /* article is not on page... */
	    (void)s_fillpage_backward(b);  /* fill backward (*must* work) */
	    *s_ptr_page_line = *s_bot_ent;  /* go to bottom of page */
	    (void)s_refillpage();	/* make sure page is full */
	    *s_ref_all = TRUE;		/* make everything redrawn... */
	    break;
	case 'N':	/* next newsgroup */
	    return(SA_NEXT);
	case 'P':	/* previous newsgroup */
	    return(SA_PRIOR);
	case 'q':	/* quit newsgroup */
	    return(SA_QUIT);
	case 'Q':	/* start reading and quit SA mode */
	    sa_in = FALSE;
	    /* FALL THROUGH */
	case '\n':
	case ' ':
	    b = sa_readmarked_elig();
	    if (b) {
		sa_clearmark(b);
		return(b);
	    }
	    /* FALL THROUGH */
	case 'r':	/* read article... */
	    /* everything needs to be refreshed... */
	    *s_ref_all = TRUE;
	    return(a);
	case 'm':	/* toggle mark on one article */
	case 'M':	/* toggle mark on thread */
	    s_rub_ptr();
	    (void)sa_art_cmd((*buf == 'M'),SA_MARK,a);
	    if (!sa_mark_stay) {
		/* go to next art on page or top of page if at bottom */
		if (*s_ptr_page_line<(*s_bot_ent)) { /* more on page */
		    *s_ptr_page_line +=1;
		}
		else
		    s_go_top_page();	/* wrap around to top */
	    }
	    break;
	case 's':	/* toggle select1 on one article */
	case 'S':	/* toggle select1 on a thread */
	    if (!sa_mode_zoom)
		s_rub_ptr();
	    (void)sa_art_cmd((*buf == 'S'),SA_SELECT,a);
	    /* if in zoom mode, selection will remove article(s) from the
	     * page, so that moving the cursor down is unnecessary
	     */
	    if (!sa_mark_stay && !sa_mode_zoom) {
		/* go to next art on page or top of page if at bottom */
		if (*s_ptr_page_line<(*s_bot_ent)) { /* more on page */
		    *s_ptr_page_line +=1;
		}
		else
		    s_go_top_page();	/* wrap around to top */
	    }
	    break;
	default:
	    s_beep();
	    return(0);
    } /* switch */
    return(0);
}

/* sa_art_cmd primitive: actually does work on an article */
void
sa_art_cmd_prim(cmd,a)
int cmd;
ART_NUM a;
{
    /* reindent this sometime */
/* do more onpage status refreshes when in unread+read mode? */
	switch (cmd) {
	    case SA_KILL_MARKED:
		if (sa_marked(a)) {
		    sa_clearmark(a);
		    oneless(a);
		}
		break;
	    case SA_KILL_UNMARKED:
		if (sa_marked(a))
		    break;		/* end case early */
		oneless(a);
		break;
	    case SA_KILL:		/* junk this article */
		sa_clearmark(a);	/* clearing should be fast */
		oneless(a);
		break;
	    case SA_MARK:		/* mark this article */
		if (sa_marked(a))
		    sa_clearmark(a);
		else
		    sa_mark(a);
		s_ref_status_onpage(a);
		break;
	    case SA_SELECT:		/* select this article */
		if (sa_selected1(a)) {
		    sa_clearselect1(a);
		    if (sa_mode_zoom)
			*s_refill = TRUE;	/* article is now ineligible */
		}
		else
		    sa_select1(a);
		s_ref_status_onpage(a);
		break;
	} /* switch */
}

/* return value is unused for now, but may be later... */
/* note: refilling after a kill is the caller's responsibility */
int
sa_art_cmd(multiple,cmd,a)
int multiple;		/* follow the thread? */
int cmd;		/* what to do */
ART_NUM a;		/* article # to affect or start with */
{
    ART_NUM b;

    sa_art_cmd_prim(cmd,a);	/* do the first article */
    if (!multiple)
	return(0);		/* no problem... */
    b = a;
#ifdef HCACHE
    hc_setspin(TRUE);
#endif
    while ((b=sa_subj_thread_next(b)))
	/* if this is basically eligible and the same subject thread# */
	sa_art_cmd_prim(cmd,b);
#ifdef HCACHE
    hc_setspin(FALSE);
#endif
    return(0);
}

/* finishes a command on the bottom line... */
/* returns TRUE if command entered, FALSE if wiped out... */
int
sa_finish_cmd(string)
char *string;
{
    *s_ref_bot = TRUE;			/* will need to be refreshed */
    s_goxy(0,LINES-*s_bot_lines);	/* go to bottom bar */
    erase_eol();			/* erase to end of line */
    s_goxy(0,LINES-*s_bot_lines);	/* go (back?) to bottom bar */
    if (string && *string) {
	printf("%s",string);
	fflush(stdout);
    }
    buf[1] = FINISHCMD;
    return(finish_command(FALSE));	/* do not echo newline */
}

ART_NUM
sa_wrap_next_author(a)
ART_NUM a;
{
    ART_NUM b;
    char *s;

    s = (char *)sa_desc_author(a,20);	/* 20 characters should be enough */
#ifdef HCACHE
    hc_setspin(TRUE);
#endif
    for (b=sa_next_elig(a);b;b=sa_next_elig(b))
#ifdef HAVE_STRSTR
	if (strstr(get_from_line(b),s))
#else
	if (trn_strstr(get_from_line(b),s))
#endif
	    break;	/* out of the for loop */
    if (b) {	/* found it */
#ifdef HCACHE
	hc_setspin(FALSE);
#endif
	return(b);
    }
    /* search from first article (maybe return original art) */
    b = sa_first();
    if (!sa_eligible(b))
	b = sa_next_elig(b);
    for (;b;b=sa_next_elig(b))
#ifdef HAVE_STRSTR
	if (strstr(get_from_line(b),s))
#else
	if (trn_strstr(get_from_line(b),s))
#endif
	    break;	/* out of the for loop */
#ifdef HCACHE
    hc_setspin(FALSE);
#endif
    return(b);
}
#endif /* SCAN */
