/* This file Copyright 1992 by Clifford A. Adams */
/* sadesc.c
 *
 */

#include "EXTERN.h"
#include "common.h"
#ifdef SCAN
#include "bits.h"
#include "head.h"
#ifdef USETHREADS
/* currently used for fast author fetch when group is threaded */
#include "ngdata.h"
#include "term.h"	/* for standout */
#include "threads.h"
#include "rthreads.h"
#endif
#include "scanart.h"
#include "samain.h"
#include "sacmd.h"	/* for sa_wrap_next_author() */
#include "sadisp.h"
#include "sathread.h"
#include "scan.h"
#ifdef SCORE
#include "score.h"
#endif
#include "INTERN.h"
#include "sadesc.h"


/* returns statchars in temp space... */
char *
sa_get_statchars(a,line)
ART_NUM a;
int line;		/* which status line (1 = first) */
{
    ART_NUM b;
    static char char_buf[256];

/* Debug */
/* printf("entry: sa_get_statchars(%d,%d)\n",(int)a,line) FLUSH; /* */

    switch (line) {
	case 1:
	    strcpy(char_buf,".....");
	    if (sa_marked(a))
		char_buf[4] = 'x';
	    if (sa_selected1(a))
		char_buf[3] = '+';
#ifdef SCORE
	    if (scoreflags[AOFFSET(a)] & SFLAG_AUTHOR)
		char_buf[1] = 'x';
	    else {
		if (sa_auth_showstat && (b = sa_wrap_next_author(a)) && (b!=a))
		    char_buf[1] = '*';
		else
		    char_buf[1] = 'o';
	    }
#endif
	    if (was_read(a))
		char_buf[0] = '-';
	    else
		char_buf[0] = '+';
	    break;
	default:
	    strcpy(char_buf,"     ");
	    break;
    } /* switch */

    return(char_buf);
}

char *
sa_desc_subject(a)
ART_NUM a;
{
    char *s,*s1;
    static char sa_subj_buf[256];
#ifdef USETHREADS
/* Still testing this modification */
    if (ThreadedGroup && sa_speed_over_acc
#ifdef HCACHE
	&& !(hcache && hcache[SUBJ_LINE] && HC_ELEM(SUBJ_LINE,AOFFSET(a)))
#endif
       ) {
	int subj1,subj2;	/* for subject comparison */

	find_article(a);
	if (p_art) {	/* we could find the article */
	    subj1 = (int)(p_art->subject);
	    if ((subj1<0) || (!*subject_ptrs[subj1])) {
		sprintf(sa_subj_buf,"<No Subject>");
		return(sa_subj_buf);
	    }
	    strncpy(sa_subj_buf+2,subject_ptrs[subj1],250);
	    sa_subj_buf[0] = '>';
	    sa_subj_buf[1] = ' ';	/* set up in case reply */
	    /* add a ">" to the beginning if it is (probably) a reply */
#ifdef UNDEF
	    if (p_art->parent)
		return(sa_subj_buf);
	    else
		return(sa_subj_buf+2);
#endif /* */
	    if (p_art->parent) {
		/* try to see if it has the same subject */
		p_art += p_art->parent;
		if (p_art) {
		    if (p_art->subject>=0) {
			subj2 = (int)(p_art->subject);
			if (subj1!=subj2) {
			    return(sa_subj_buf+2);	/* not a reply */
			}
			return(sa_subj_buf);  /* reply (or no parent subj) */
		    }
		    /* parent article has no subject */
		    return(sa_subj_buf);   /* assume reply */
		} /* if (p_art) */
	    } /* if (p_art->parent) */
	    /* no parent article */
	    /* later look for siblings with same subject? */
	    return(sa_subj_buf+2);	/* not reply */
	} /* if (p_art) */
    }	/* if Threaded but not article subject not cached */
#endif /* USETHREADS */
    s1 = s = fetchsubj(a,FALSE,FALSE);

    if (!s || !*s) {
	sprintf(sa_subj_buf,"(no subject)");
	return(sa_subj_buf);
    }
    strcpy(sa_subj_buf,s);
    s1 = sa_subj_buf;
    if ((tolower(*s1++) == 'r') && (tolower(*s1++) == 'e') && (*s1 ==':')) {
	*s1 = '>';		/* more cosmetic "Re:" */
	return(s1);
    }
    return(sa_subj_buf);
}

/* returns author's name up to length characters in a static array */
char *
sa_desc_author(a,length)
ART_NUM a;
int length;
{
    char *s,*s1,*s2,*s3;
    static char sa_auth_buf[256];

    /* small speedup--get the author's name from thread data
     * (if group is threaded and the art is not in the hcache)
     */
/* should this be dependent on sa_use_trndb?  I don't think so now
 * since author data is only used for display.
 */
#ifdef USETHREADS
    if (ThreadedGroup
#ifdef HCACHE
	&& !(hcache && hcache[FROM_LINE] && HC_ELEM(FROM_LINE,AOFFSET(a)))
#endif
	) { /* if threaded and no hcache then grab fast from DB */
	find_article(a);
	if (p_art) {	/* we could find the article */
	    s = author_ptrs[p_art->author];
	    strncpy(sa_auth_buf,s,255);
	    return(sa_auth_buf);
	}
	else {
	    sprintf(sa_auth_buf,"<No Author>");
	    return(sa_auth_buf);
	}
    }
#endif

    s = get_from_line(a);	/* buffered, not allocated return */
    if (!s || !*s)		/* no author, implies unavail... */
	sprintf(sa_auth_buf,"<No Author>");
    else
	strcpy(sa_auth_buf,s);
    s1 = s = sa_auth_buf;
    while (*s1 && (*s1++ != '('))
	;	/* EMPTY */
    if (*s1) {		/* opening of author */
	s2 = s1;
	while (*s2 && (*s2 != ')'))
	    s2++;
	if (*s2) {
	    /* check for not just whitespace */
	    for (s3 = s1; ((s3 != s2) && (*s3 == ' ')) ; s3++)
		;	/* EMPTY */
	    if (s3 != s2) {	/* there were non-space characters */
		*s2 = '\0';
		s = s1;		/* use the short name for the author */
	    }
	}
    } /* end of author name test */
    /* shorten string to length */
    s[length] = '\0';
    return(s);
}

char *
padspaces(s,size)
char *s;
int size;
{
    int len;
    char *s1;
    static char sa_pad_buf[1024];

    len = strlen(s);
    if (len>=size)
	return(s);
    strcpy(sa_pad_buf,s);
    s = sa_pad_buf;
    for (s1=s+len;s1<s+size;s1++)
	*s1 = ' ';
    *s1 = '\0';
    return(s);
}
	

/* NOTE: should redesign later for the "menu" style... */
char *
sa_get_desc(a,line)
ART_NUM a;
int line;
{
    static char desc_buf[1024];
    char *s;
    bool use_standout;	/* if TRUE, use stdout on line */

    use_standout = FALSE;
    switch (line) {
	case 1:
	    desc_buf[0] = '\0';	/* initialize the buffer */
	    if (sa_mode_desc_artnum) {
		sprintf(sa_buf,"%6d ",a);
		 strcat(desc_buf,sa_buf);
	    }
#ifdef SCORE
	    if (sc_initialized && sa_mode_desc_score) {
		/* we'd like the score now */
		sprintf(sa_buf,"[%4d] ",sc_score_art(a,TRUE));
		strcat(desc_buf,sa_buf);
	    }
#endif /* SCORE */
	    if (sa_mode_desc_threadcount) {
		sprintf(sa_buf,"(%3d) ",subj_thread_count(a));
		strcat(desc_buf,sa_buf);
	    }
	    if (sa_mode_desc_author) {
		sprintf(sa_buf,"%s ",padspaces(sa_desc_author(a,16),16));
		strcat(desc_buf,sa_buf);
	    }
	    if (sa_mode_desc_subject) {
		sprintf(sa_buf,"%s",sa_desc_subject(a));
		strcat(desc_buf,sa_buf);
	    }
	break;
	case 2:	/* summary line (test) */
	    s = fetchlines(a,SUMRY_LINE);
	    if (s && *s) { /* we really have one */
		int i;		/* number of spaces to indent */
		char *s2;	/* for indenting */

/* uncomment the following line to use standout mode */
/*		use_standout = TRUE; /* */
		i = 0;
		/* if variable widths used later, use them */
		if (sa_mode_desc_artnum)
		    i += 7;
#ifdef SCORE
		if (sc_initialized && sa_mode_desc_score)
		    i += 7;
#endif
		if (sa_mode_desc_threadcount)
		    i += 6;
		s2 = desc_buf;
		while (i--)
		    *s2++ = ' ';
#ifdef HAVETERMLIB
		if (use_standout)
		    sprintf(s2,"Summary: %s%s",SO,s);
		else
#endif
		    sprintf(s2,"Summary: %s",s);
		break;
	    }
	/* otherwise, we might have had a keyword */
	/* FALL THROUGH */
	case 3:	/* Keywords (test) */
	    s = fetchlines(a,KEYW_LINE);
	    if (s && *s) { /* we really have one */
		int i;		/* number of spaces to indent */
		char *s2;	/* for indenting */

/* uncomment the following line to use standout mode */
/*		use_standout = TRUE; /* */
		i = 0;
		/* if variable widths used later, use them */
		if (sa_mode_desc_artnum)
		    i += 7;
#ifdef SCORE
		if (sc_initialized && sa_mode_desc_score)
		    i += 7;
#endif
		if (sa_mode_desc_threadcount)
		    i += 6;
		s2 = desc_buf;
		while (i--)
		    *s2++ = ' ';
#ifdef HAVETERMLIB
		if (use_standout)
		    sprintf(s2,"Keys: %s%s",SO,s);
		else
#endif
		    sprintf(s2,"Keys: %s",s);
		break;
	    }
	default:	/* no line I know of */
	    /* later return NULL */
	    sprintf(desc_buf,"Article %d: Nonimplemented Description LINE",a);
	    break;
    } /* switch (line) */

    desc_buf[*s_desc_cols] = '\0';	/* make sure it's not too long */
#ifdef HAVETERMLIB
    if (use_standout)
	strcat(desc_buf,SE);	/* end standout mode */
#endif
    /* take out tabs (replace with one space) */
    for (s=desc_buf;*s;s++)
	if (*s == '\t')
	    *s = ' ';
    return(desc_buf);
}

/* returns # of lines the article occupies in total... */
int
sa_art_lines(a)
ART_NUM a;
{
    char *s;
    int num = 1;

    if (sa_mode_desc_summary) {
	s = fetchlines(a,SUMRY_LINE);
	if (s && *s)
	    num++;	/* just a test */
	if (s)
	    free(s);
    }
    if (sa_mode_desc_keyw) {
	s = fetchlines(a,KEYW_LINE);
	if (s && *s)
	    num++;	/* just a test */
	if (s)
	    free(s);
    }
    return(num);
}

#endif /* SCAN */
