/* This file Copyright 1992 by Clifford A. Adams */
/* samain.c
 *
 * main working routines  (may change later)
 */

#include "EXTERN.h"
#include "common.h"
#ifdef SCAN
#include "bits.h" /* variable absfirst, for the AOFFSET macro... */
#include "head.h" /* for fetching misc header lines... */
#include "term.h" /* for LINES */
#include "util.h" /* for savestr */
#include "intrp.h"	/* for filexp */
#ifdef SCORE
#include "score.h"
#endif
#include "scan.h"
#include "scmd.h"
#include "sdisp.h"
#include "smisc.h"
#include "spage.h"
#include "scanart.h"
#include "samisc.h"
#include "sacmd.h"
#include "sadesc.h"
#include "sadisp.h"
#include "sathread.h"
#include "INTERN.h"
#include "samain.h"

static int sa_scan_context INIT(0);

void
sa_init_context()
{
    if (sa_context_init)
	return;		/* already initialized */
    if (sa_scan_context==0) {
	sa_scan_context = s_new_context(S_ART);
    }
    s_change_context(sa_scan_context);
}

void
sa_init()
{
    sa_init_context();
    if (s_initscreen())		/* If not able to init screen...*/
	return;				/* ...most likely dumb terminal */
    if (!sa_initarts())			/* init article array(s) */
	return;				/* ... no articles */
    init_subj_threads();		/*      subject threads */
    sa_initmode();			/*      mode differences */
    sa_initialized = TRUE;		/* all went well... */
}

void
sa_cleanmain()
{
    free(sa_per_art);

    /* free the subject thread strings */
    sa_clean_threads();
    sa_mode_zoom = FALSE;	/* doesn't survive across groups */
    s_delete_context(sa_scan_context);
    sa_context_init = FALSE;
    sa_scan_context = 0;
    sa_in = FALSE;
}

void
sa_growarts(oldlast,last)
ART_NUM oldlast,last;
{
    int i;

    printf("\nGrowing newsgroup: sa_morearts(%d,%d)",oldlast,last);
    fflush(stdout);
    sa_per_art = (SA_PER *)saferealloc((char*)sa_per_art,
	(AOFFSET(last)+1)*sizeof(SA_PER));
    for (i=AOFFSET(oldlast+1) ; i<=AOFFSET(last) ; i++) {
	sa_per_art[i].subj_thread_num = 0;
	sa_per_art[i].sa_art_status = (char)0;
    }
/* set the refill flag?  Refill? */
    printf("Done.\n");
    fflush(stdout);
}

/* to be called sometime after initctl in bits.h (called from there?) */
bool_int
sa_initarts()
{
    int i;

    if (lastart==0)	/* no articles at all */
	return(FALSE);

    sa_per_art = (SA_PER*)safemalloc((AOFFSET(lastart)+1)*sizeof(SA_PER));
    for (i=0 ; i<=AOFFSET(lastart) ; i++) {
	sa_per_art[i].subj_thread_num = 0;
	sa_per_art[i].sa_art_status = (char)0;
    }
    return(TRUE);
}

/* note: initscreen must be called before (for scr_width) */
void
sa_initmode()
{
    char *s;
    unsigned int i;

    /* set up screen sizes */
    sa_set_screen();

    if (!sa_never_initialized)
	return;		/* only initialize the first time */

    /* beginning to use rc files now */
    /* allow SSDISPLAY to be used in case SADISPLAY is not set */
    s = getval("SADISPLAY","xxx");
    if (strEQ(s,"xxx"))
	s = getval("SSDISPLAY","16");	/* default to only subject */
    i = atoi(s);
    sa_mode_desc_artnum = (i & 1);
#ifdef SCORE
    sa_mode_desc_score = (i & 2);
#endif
    sa_mode_desc_threadcount = (i & 4);
    sa_mode_desc_author = (i & 8);
    sa_mode_desc_subject = (i & 16);
    sa_mode_desc_summary = (i & 32);
    sa_mode_desc_keyw = (i & 64);

    /* allow SSDISPLAY to be used in case SADISPLAY is not set */
    s = getval("SAMODE","xxx");
    if (strEQ(s,"xxx"))
	s = getval("SSMODE","8");	/* default to follow */
    i = atoi(s);
#ifdef SCORE
    score_newfirst = (i & 2);
    if (i & 1)
	sa_mode_order = 2;
    else
#endif
	sa_mode_order = 1;
    sa_mode_read_elig = (i & 4);
    sa_follow = (i & 8);
    sa_mode_fold = (i & 16);
    sa_mode_zoom = (i & 32);
    sa_mode_vi = (i & 64);
    /* set the scan-global vi mode flag */
    if (sa_mode_vi)
	s_mode_vi = 2;
    else
	s_mode_vi = 1;
    sa_mark_stay = (i & 128);
    /* CAA: shift the bits over by one byte since the & operator
     * doesn't seem to work for higher bits.  (I thought it should,
     * but obviously the compiler has a different idea. :)
     */
    i = i>>8;	/* divide by 256 */
#ifdef USETHREADS
    sa_speed_over_acc = (i & 1);
#endif
}

/* not needed, but might be later.  Probably delete before beta. */
void
sa_cleanarts()
{
}

int
sa_mainloop()
{
    int i;

#ifdef SCORE
    /* if, for whatever reason, we are in score mode but scoring was not
     * initialized, then scoring must have failed.  Will need a variable
     * in score.[ch] set when scoring *failed*.  For now, just go back to
     * arrival order.  (Failure variable needed so that we could try again
     * or decide to disallow score operations.)
     */
    if ((!sc_initialized) && (sa_mode_order == 2))
	sa_mode_order = 1;	/* arrival order */
#endif
    /* redraw it *all* */
    *s_ref_all = TRUE;
    if ((*s_top_ent<absfirst) || (*s_top_ent>lastart))
	*s_top_ent = s_first();
    i = s_fillpage();
    if ((i==-1) || (i == 0)) {
	/* for now just quit if no page could be filled. */
	return(SA_QUIT);
    }
    i = s_cmdloop();
    if (i>0) {
	sa_art = i;
	return(SA_NORM);
    }
    /* something else (quit, return, etc...) */
    return(i);
}

void
sa_lookahead()
{
#ifdef PENDING
    ART_NUM a;		/* used in #ifdef'ed code */

#ifdef SCORE
    sc_lookahead(TRUE,FALSE);		/* do resorting now... */
#else /* !SCORE */
    /* look forward from the last article on the page... */
    a = (ART_NUM)page_ents[*s_bot_ent].entnum;
    while ((!input_pending()) && (a)) {
	(void)subj_thread(a);
	if (sa_mode_desc_author)	/* lookahead for authors too */
	    (void)get_from_line(a);
	a = sa_next_elig(a);
    }
#endif /* SCORE */
#endif /* PENDING */
}

/* returns first marked article #
 * returns 0 if no marked articles */

ART_NUM
sa_readmarked_elig()
{
    ART_NUM a;

    /* note: sa_marked is fast, but sa_next is slow.
     * as this is commonly called when there are *no* marked articles,
     * first check to see if there are any marked ones...
     */
    for (a=absfirst;a<=lastart;a++)
	if sa_marked(a)
	    break;
    if (a>lastart)	/* no marked ones */
	return((ART_NUM)0);
    a = sa_first();
    if (!a)
	return((ART_NUM)0);
    if (!sa_basic_elig(a))
	a=sa_next(a);
    for (;a;a = sa_next(a))
	if (sa_marked(a))
	    return(a);
    /* note: this is valid since the marked one(s) might not be elig. */
    return((ART_NUM)0);
}

#ifdef UNDEF
void
sa_main_resize_win()
{
    int oldheight,i;

    oldheight = scr_height;
    if (s_initscreen())		/* couldn't initialize */
	return;			/* bad error, but what should we do? */
    /* perhaps later call a routine in sapage.c */
    if (scr_height<oldheight) {	/* screen shrunk */
	int page_lines;	
	ART_NUM a;

	page_lines = scr_height - *s_top_lines - *s_bot_lines;
	for (i=0;i<=*s_bot_ent;i++) {
	    a = page_ents[i].entnum;
	    if ((page_ents[i].start_line>page_lines) ||
		((page_ents[i].start_line+s_ent_lines(a))>page_lines))
		    *s_bot_ent = i-1;	/* chop page */
	}
    /* Now, suppose that the pointer position is off the page... */
    if (*s_ptr_page_line>*s_bot_ent)
	*s_ptr_page_line = *s_bot_ent;
    } else if (scr_height>oldheight)
	*s_refill = TRUE;	/* could use a refill */
    sa_initmode();		/* consider splitting later? */
    *s_ref_all = TRUE;		/* need a total refresh */
}
#endif
#endif /* SCAN */
