/* This file Copyright 1992 by Clifford A. Adams */
/* samisc.c
 *
 * lower-level routines
 */

#include "EXTERN.h"
#include "common.h"
#ifdef SCAN
#include "bits.h"		/* AOFFSET (absfirst) */
#include "final.h"		/* assert's sig_catcher() */
#include "term.h"		/* for "backspace" */
#include "util.h"
#include "scanart.h"
#include "samain.h"
#include "sathread.h"
#include "sadesc.h"		/* TEMPORARY(?) */
#ifdef USETHREADS		/* for sa_selthreads */
#include "threads.h"
#include "rthreads.h"
#include "rn.h"
#include "rcstuff.h"
#include "ng.h"			/* for "art" */
#endif /* USETHREADS */
#include "ngdata.h"		/* several */
#include "head.h"		/* fetchsubj */
#ifdef SCORE
#include "score.h"
#endif
#include "INTERN.h"
#include "samisc.h"

int
check_article(a)
ART_NUM a;
{
    if ((a<absfirst) || (a>lastart)) {
	printf("\nArticle %d out of range\n",a) FLUSH;
	return(FALSE);
    }
    return(TRUE);
}

#ifdef UNDEF	/* do this with final systems */
#define check_article(a) ((a>=absfirst) && (a<=lastart))
#endif

/* ignoring "Fold" (or later recursive) mode(s), is this article eligible? */
bool_int
sa_basic_elig(a)
ART_NUM a;
{
    if(!check_article(a)) {
	printf("sa_basic_elig: article out of range\n");
	assert(check_article(a));	/* should fail */
	return(FALSE);
    }
    /* "run the gauntlet" style (:-) */
    if ((!sa_mode_read_elig) && was_read(a))
	return(FALSE);
    if (sa_mode_zoom && !sa_selected1(a))
	return(FALSE);
    /* now just check availability */
    if (is_unavailable(a)) {
	if (!was_read(a))
	    oneless(a);
	return(FALSE);		/* don't try positively unavailable */
    }
    if (is_available(a))
	return(TRUE);
#ifdef USETHREADS
    if (sa_speed_over_acc) {
/* if any fast checks can be figured out, do them here */
	return(TRUE);  /* reasonable assumption */
    }
#endif
    /* not explicitly available or unavailable */
    (void)fetchsubj(a,FALSE,FALSE);	/* now we will know... */
    if (is_unavailable(a)) {
	if (!was_read(a))
	    oneless(a);
	return(FALSE);		/* don't try positively unavailable */
    }
    return(TRUE);	/* passed all tests */
}

bool
sa_eligible(a)
ART_NUM a;
{

    if(!check_article(a)) {
	printf("sa_eligible: article out of range\n");
	assert(check_article(a));	/* should fail */
	return(FALSE);
    }
    if (!sa_basic_elig(a))
	return(FALSE);		/* must always be basic-eligible */
    if (!sa_mode_fold)
	return(TRUE);		/* just use basic-eligible */
    else {
#ifdef USENOV
	/* threadfold here later */
#endif
	if (sa_subj_thread_prev(a))
	    return(FALSE);	/* there was an earlier match */
	return(TRUE);		/* no prior matches */
    }
}

/* given an article #, returns "previous" basic_elig. article # */
/* returns 0 if no previous article */
ART_NUM
sa_prev(a)
ART_NUM a;
{
    ART_NUM b;
    int i;

    if (a<absfirst || a>lastart)	/* out of range */
	return(0);		/* should we complain? */
    switch (sa_mode_order) {
#ifdef SCORE
	case 2:		/* score descending */
	    b = a;
	    while ((b = sc_art_higher_score(b)) && !sa_basic_elig(b))
		;	/* EMPTY */
	    return(b);
#endif
	case 1: 	/* arrival mode */
	default:
	    b = a;
	    b--;
	  while(TRUE) {
 	    for (i=0;i<3;i++) {
		if ((b<=absfirst) ||
		    ((b<firstart) && !sa_mode_read_elig))
			return(0);
		if (sa_basic_elig(b))
		    return(b);
		if (!is_unavailable(b))
		    i = 0;	/* reset counter */
		b--;
	    }
	    /* three unavail...  try something else... */
	    b = get_prev_avail(b+1);	/* will mark and set avail... */
	  } /* while */
    }
}

/* given an article #, returns "next" basic_elig article */
/* returns 0 if no next article */
ART_NUM
sa_next(a)
ART_NUM a;
{
    ART_NUM b;
    int i;

    if (a<absfirst || a>lastart)	/* out of range */
	return(0);		/* should we complain? */
    switch (sa_mode_order) {
#ifdef SCORE
	case 2:		/* score descending */
	    b = a;
	    while ((b = sc_art_lower_score(b)) && !sa_basic_elig(b))
		;	/* EMPTY */
	    return(b);
#endif
	case 1:		/* arrival mode */
	default:
	    b = a;
	    b++;
	  while(TRUE) {
 	    for (i=0;i<3;i++) {
		if (b>lastart)
		    return(0);
		if (sa_basic_elig(b))
		    return(b);
		if (!is_unavailable(b))
		    i = 0;	/* reset counter */
		b++;
	    }
	    /* three unavail...  try something else... */
	    b = get_next_avail(b-1);	/* will mark and set avail... */
	  } /* while */
    }
}

/* given an article, returns previous eligible article */
/* returns 0 if no previous eligible article */
ART_NUM
sa_prev_elig(a)
ART_NUM a;
{
    while((a = sa_prev(a)))
	if (sa_eligible(a))
	    return a;
    return (ART_NUM)0;
}

/* given an article, returns next eligible article */
/* returns 0 if no next eligible article */
ART_NUM
sa_next_elig(a)
ART_NUM a;
{
    while((a = sa_next(a)))
	if (sa_eligible(a))
	    return a;
    return (ART_NUM)0;
}

/* returns "first" article# ("first" depends on mode) */
/* note: not necessarily an eligible article: use sa_next_elig if not */
ART_NUM
sa_first()
{
    if (absfirst>lastart)		/* no articles *exist* */
	return(0);
    switch (sa_mode_order) {
#ifdef SCORE
	case 2:		/* score descending */
	    return(sc_art_highest_score());
#endif
	case 1:		/* arrival mode */
	default:
	    if (sa_mode_read_elig)
		return absfirst;
	    else {
		if (firstart<=lastart)
		    return firstart;
		else
		    return(lastart);
	    }
    } /* switch */
}

/* returns "last" article# ("last" depends on mode) */
/* note: not necessarily an eligible article: use sa_prev_elig if not */
ART_NUM
sa_last()
{
    if (absfirst>lastart)		/* no articles *exist* */
	return(0);
    switch (sa_mode_order) {
#ifdef SCORE
	case 2:		/* score descending */
	    return(sc_art_lowest_score());
#endif
	case 1:		/* arrival mode */
	default:
	    return lastart;
    }
}

#ifdef USETHREADS
/* select1 the articles picked in the TRN thread selector */
void
sa_selthreads()
{
    PACKED_ARTICLE *root_limit;
    register int r;

    bool want_unread = unread_selector;

/* DEBUG */
/*  printf("entry: sa_selthreads\n") FLUSH; /* */

    {	/* clear any old selections */
	ART_NUM i;

	for (i=0;i<=AOFFSET(lastart);i++)
	    sa_per_art[i].sa_art_status =
	      (sa_per_art[i].sa_art_status & 0xfd);
    }

    /* Copied and edited from use_selected in ngstuff.c */

    /* Multiple commands and commands that operate on individual articles
    ** use the article loop.
    */

#ifdef SCORE
#ifdef HCACHE
    if (sa_mode_order == 2)	/* in score order */
	hc_setspin(TRUE);	/* scoring may take time... */
#endif
#endif
    for (r = 0; r < total.root; r++) {
	if (selected_roots[r]) {
		p_art = p_articles + p_roots[r].articles;
		root_limit = upper_limit( p_art, 0 );
		for (; p_art < root_limit; p_art++) {
		    art = p_art->num;
		    if (p_art->subject != -1
		     && (!was_read(art) ^ want_unread)) {
			/* this was a trn-thread selected article */
			sa_select1(art);
#ifdef SCORE
    /* if scoring, make sure that this article is scored... */
			if (sa_mode_order == 2)	/* score order */
			    sc_score_art(art,FALSE);
#endif
		    }
		    if (p_art == Nullart)
			break;
		}/* for all articles */
	}/* if selected */
    }/* for all threads */
#ifdef SCORE
#ifdef HCACHE
    if (sa_mode_order == 2) {	/* in score order */
	hc_setspin(FALSE);
	sc_sort();
    }
#endif
#endif
}
#endif /* USETHREADS */

void
sa_keypause()
{
    printf("\nsa:Press a key to continue\n") FLUSH;
/* Note: this is nasty, but I don't want side effects when debugging */
    while (!input_pending())
	;
    eat_typeahead();
}
#endif /* SCAN */

int
sa_number_arts()
{
    int total;
    ART_NUM i;

    if (sa_mode_read_elig)
	i = absfirst;
    else
	i = firstart;
    total = 0;
    for (;i<=lastart;i++) {
	if (is_unavailable(i))
	    continue;
	if (was_read(i) && !sa_mode_read_elig)
	    continue;
	total++;
    }
    return(total);
}
