/* This file Copyright 1992 by Clifford A. Adams */
/* sathread.c
 *
 */

#include "EXTERN.h"
#include "common.h"
#ifdef SCAN
#include "bits.h"	/* for absfirst */
#ifdef HCACHE
#include "head.h"	/* hc_setspin() */
#endif
#include "util.h"	/* savestr() */
#include "scanart.h"
#include "sadesc.h"	/* sa_desc_subject() */
#include "samain.h"
#include "samisc.h"
#include "INTERN.h"
#include "sathread.h"

char **subj_threads;
static int num_subj_threads = 0;

void
init_subj_threads()
{
    num_subj_threads = 0;
    subj_threads = Null(char **);
}

int
subj_thread(a)
ART_NUM a;
{
    int i,num;
    char *s,*s1;

    if ((num = sa_per_art[AOFFSET(a)].subj_thread_num))  /* EQUALS */
	return(num);

    num = -1;

    s = sa_desc_subject(a);	/* note: will always return good string */
    for (i=0;i<num_subj_threads;i++) {
	s1 = s;
	if ((!strcasecmp(s,subj_threads[i])) ||
	    ((*s1++ == '>') && (*s1++ == ' ') &&
	     (!strcasecmp(s1,subj_threads[i])))) {
		num = i;
		break;
	}
    }
    if (num>-1) {		/* there was a match */
	sa_per_art[AOFFSET(a)].subj_thread_num = num+1;
	return(num+1);
    }
    else {		/* no existing match */
	num_subj_threads++;
	if (num_subj_threads==1)
	    subj_threads = (char**)safemalloc(sizeof(char*));
	else
	    subj_threads = (char**)saferealloc((char*)subj_threads,
				num_subj_threads*sizeof(char*));
	s1 = s;
	if ((*s1++ == '>') && (*s1++ == ' '))
	    s = s1;		/* skip the "> " */
	subj_threads[num_subj_threads-1] = savestr(s);
	sa_per_art[AOFFSET(a)].subj_thread_num = num_subj_threads;
	return(num_subj_threads);
    }
}

int
subj_thread_count(a)
ART_NUM a;
{
    int count;
    ART_NUM b;

    count = 1;
    b = a;

#ifdef HCACHE
    hc_setspin(TRUE);
#endif
    while ((b = sa_subj_thread_next(b)))
	if (sa_basic_elig(b))
	    count++;
#ifdef HCACHE
    hc_setspin(FALSE);
#endif
    return(count);
}

/* returns basic_elig previous subject thread */
ART_NUM
sa_subj_thread_prev(a)
ART_NUM a;
{
    int i,j;

    i = subj_thread(a);
    while ((a = sa_prev(a))) {
	if (!(j = sa_per_art[AOFFSET(a)].subj_thread_num))
	    j = subj_thread(a);
	if (i == j)
	    return(a);
    }
    return((ART_NUM)0);
}

ART_NUM
sa_subj_thread_next(a)
ART_NUM a;
{
    int i,j;

    i = subj_thread(a);
    while ((a = sa_next(a))) {
	if (!(j = sa_per_art[AOFFSET(a)].subj_thread_num))
	    j = subj_thread(a);
	if (i == j)
	    return(a);
    }
    return((ART_NUM)0);
}

void
dump_subj_threads()
{
    int i;

    putchar('\n');

    for (i=0;i<num_subj_threads;i++)
	printf("[%d] |%s|\n",i,subj_threads[i]) FLUSH;
}

void
sa_clean_threads()
{
    int i;

    if (subj_threads) {
	for (i=0;i<num_subj_threads;i++)
	    free(subj_threads[i]);
	free(subj_threads);
    }
}
#endif /* SCAN */
