/* scmd.c
 *
 * Scan command loop.
 * Does some simple commands, and passes the rest to context-specific routines.
 */

#include "EXTERN.h"
#include "common.h"
/* absfirst, lastart, oneless() */
#include "bits.h"
#include "help.h"
#include "term.h"	/* for LINES */
#include "util.h"
#include "scan.h"
#include "smisc.h"
#include "spage.h"
#include "sdisp.h"
#include "sgroup.h"
#include "sgcmd.h"	/* sg_docmd */
#include "sgmisc.h"	/* sg_lookahead() */
#include "sacmd.h"	/* sa_docmd */
#include "samain.h"	/* sa_lookahead */
#ifdef SCORE
#include "score.h"
#endif
#include "INTERN.h"
#include "scmd.h"

/* returns an entry # selected, S_QUIT, or S_ERR */
int
s_cmdloop()
{
    int i;

    /* initialization stuff for entry into s_cmdloop */
    *s_ref_all = TRUE;
    if (s_mode_vi==0) {
	/* really crude ugly hack.  Users tend to like to keep their old
	 * features though...
	 * (Backwards compatibility with ScreenScan)
	 */
	char *s;

	s = getval("SAMODE","xxx");
	if (strEQ(s,"xxx"))
	    s = getval("SSMODE","0");
	i = atoi(s);
	if (i & 64)
	    s_mode_vi = 2;
	else
	    s_mode_vi = 1;
    }
    eat_typeahead();	/* no typeahead before entry */
    while(TRUE) {
	s_refresh();
	s_place_ptr();		/* place article pointer */
	s_lookahead();		/* do something useful while waiting */
	getcmd(buf);
	eat_typeahead();	/* stay in control. */
	/* check for window resizing and refresh */
	/* if window is resized, refill and redraw */
	if (s_resized) {
	    char ch = *buf;
	    i = s_fillpage();
	    if ((i == -1) || (i == 0))	/* can't fillpage */
	        return(S_QUIT);
	    *buf = Ctl('l');
	    (void)s_docmd();
	    *buf = ch;
	    s_resized = FALSE;		/* dealt with */
	}
	i = s_docmd();
	if (i==S_NOTFOUND) {	/* command not in common set */
	    switch (s_cur_type) {
		case S_ART:
		    i = sa_docmd();
		    break;
		case S_GROUP:
		    i = sg_docmd();
		    break;
		default:
		    i = 0;	/* just keep looping */
	    }
	}
	if (i != 0)	/* either an entry # or a return code */
	    return(i);
	if (*s_refill) {
	    i = s_fillpage();
	    if ((i == -1) || (i == 0))	/* can't fillpage */
	        return(S_QUIT);
        }
	/* otherwise just keep on looping... */
    }
}

void
s_lookahead()
{
    switch (s_cur_type) {
	case S_ART:
	    sa_lookahead();
	    return;
	default:
	    sg_lookahead();
	    return;
    }
}

/* Do some simple, common Scan commands for any mode */
/* Interprets command in buf, returning 0 to continue looping or
 * a condition code (negative #s).  Responsible for setting refresh flags
 * if necessary.
 */
/* Note: VI-style movement not supported here! */
int
s_docmd()
{
    long a;		/* entry pointed to */
    long b;		/* misc. artnum */
    int i,j;		/* for misc. purposes */
    bool flag;		/* misc */

    a = page_ents[*s_ptr_page_line].entnum;
    switch(*buf) {
	case 'j':	/* vi mode */
	    if (s_mode_vi==1)
		return(S_NOTFOUND);
	    /* FALL THROUGH */
	case 'n':	/* next art */
	case ']':
	    s_rub_ptr();
	    if (*s_ptr_page_line<(*s_bot_ent)) { /* more on page... */
		*s_ptr_page_line +=1;
	    }
	    else {
		if (!s_next_elig(page_ents[*s_bot_ent].entnum)) {
		    s_beep();
		    break;
		}
		s_go_next_page();	/* will jump to top too... */
	    }
	    break;
	case 'k':	/* vi mode */
	    if (s_mode_vi==1)
		return(S_NOTFOUND);
	    /* FALL THROUGH */
	case 'p':	/* previous art */
	case '[':
	    s_rub_ptr();
	    if (*s_ptr_page_line>0) { /* more on page... */
		*s_ptr_page_line = *s_ptr_page_line - 1;
	    }
	    else {
		if (s_prev_elig(page_ents[0].entnum)) {
		    s_go_prev_page();
		    *s_ptr_page_line = *s_bot_ent; /* go to page bot. */
		} else
		    s_beep();
	    }
	    break;
	case 't':	/* top of page */
	    s_rub_ptr();
	    s_go_top_page();
	    break;
	case 'b':	/* bottom of page */
	    s_rub_ptr();
	    s_go_bot_page();
	    break;
	case '>':	/* next page */
	    s_rub_ptr();
	    a = s_next_elig(page_ents[*s_bot_ent].entnum);
	    if (!a) {		/* at end of articles */
		s_beep();
		break;
	    }
	    s_go_next_page();		/* will beep if no next page */
	    break;
	case '<':	/* previous page */
	    s_rub_ptr();
	    if (!s_prev_elig(page_ents[0].entnum)) {
		s_beep();
		break;
	    }
	    s_go_prev_page();		/* will beep if no prior page */
	    break;
	case 'T':	/* top of ents */
	case '^':
	    s_rub_ptr();
	    flag = s_go_top_ents();
	    if (!flag)		/* failure */
		return(S_QUIT);
	    break;
	case 'B':	/* bottom of ents */
	case '$':
	    s_rub_ptr();
	    flag = s_go_bot_ents();
	    if (!flag)
		return(S_QUIT);
	    break;
	case Ctl('r'):	/* refresh screen */
	case Ctl('l'):
	    *s_ref_all = TRUE;
	    break;
	case 'h':	/* help */

	    *s_ref_all = TRUE;			/* help uses whole screen */
	    s_goxy(0,LINES-*s_bot_lines);	/* go to bottom bar */
	    erase_eol();			/* erase to end of line */
	    s_goxy(0,LINES-*s_bot_lines);	/* go (back?) to bottom bar */

	    /* any commands typed during help are unused. (might change) */
	    switch (s_cur_type) {
		case S_ART:
		    (void)help_scanart();
		    break;
		case S_GROUP:
		    (void)help_scangroup();
		    break;
		default:
		    printf("No help available for this mode (yet).\n") FLUSH;
		    printf("Press any key to continue.\n");
		    break;
	    }
	    (void)get_anything();
	    break;
	default:
	    return(S_NOTFOUND);		/* not one of the simple commands */
    } /* switch */
    return(0);		/* keep on looping! */
}
