/* This file Copyright 1992 by Clifford A. Adams */
/* score.c
 *
 */

/* note: dependent on AVAILBITS feature. */

#include "EXTERN.h"
#include "common.h"
#ifdef SCORE
/* if SCORE is undefined, no code should be compiled */
/* sort the following includes later */
#include "bits.h"
#include "final.h"		/* int_count */
#include "util.h"		/* several */
#include "term.h"		/* input_pending() */
#include "head.h"		/* get_from_line */
#include "artio.h"		/* for openart var.*/
#include "ng.h"			/* art */
#include "rn.h"			/* ngname */
#include "search.h"		/* for regex */
#include "intrp.h"		/* for filexp */
#ifdef SCAN
#include "scan.h"
#include "scanart.h"
#include "samain.h"	/* for sa_mode_read_elig and more */
#include "samisc.h"	/* for sa_basic_elig() */
#endif
#ifdef NEWSFILTER
#include "filtpipe.h"
#endif
#include "INTERN.h"
#include "score.h"
#include "EXTERN.h"
#include "scorefile.h"
#include "INTERN.h"	/* to allow initialization */

/* undo an aoffset */
/* consider moving to common.h (even though no other files use it (yet?)) */
#define UNAOFFSET(x) ((x)+absfirst)

/* CONSIDER REMOVING THE "_indx" FROM THESE. */
/* minimum indexes that are valid */
static int max_indx;	/* maximum one with a score...*/
static int max_sorted_indx;	/* maximum that is sorted... */

static bool sc_announce_elig INIT(FALSE); /* print '!' when elig. scored... */
static bool sc_rescoring INIT(FALSE);	/* are we rescoring now? */

static int sc_loaded_count INIT(0);	/* how many articles were loaded? */

void
sc_init(pend_wait)
bool_int pend_wait;	/* if true, enter pending mode when scoring... */
{
    int i;
    ART_NUM a;

    if (lastart==0) {
	printf("No articles exist to be scored.\n") FLUSH;
	return;
    }
#ifdef RUNBACK
    if (use_runback)
	pend_wait = TRUE;	/* wait until all are scored */
#endif
    /* set up from environment variables */
    i = atoi(getval("SCOREMODE","3"));
    if (i>=0) {
#ifdef NICEBG
	sc_mode_nicebg = (i & 1);	/* nicebg: default on */
#endif
	sf_verbose = (i & 2);		/* verbose readin: default on */
    }

/* misc: consider the relationships between SCORE and SA.  Should one
 * be able to initialize the other?  How much can they depend on each
 * other when both are #defined?   This is one question to ask in a later
 * redesign or porting these systems to other newsreaders. */

    kill_thresh_active = FALSE;  /* kill thresholds are generic */
    sc_savescores = FALSE;
    sc_savefile = Nullch;

    sf_init();	/* initialize the scorefile code */
#ifdef RUNBACK
    if (use_runback && !sc_savescores)	{ /* no point in non-saved groups */
	sf_clean();
	return;
    }
#endif
    /* Initialize scoring arrays.  The ascore[art]list pointers are used
     * so that the heapsort has a 1-based array (ascore[art]list), while
     * the rest of STRN gets the usual 0-basis.  The arrays are:
     *  ascorelist: allocation pointer for scorelist. (1-based data)
     *  ascoreartlist: allocation pointer for scoreartlist. (1-based data)
     *  scorelist: scores sorted in nondecreasing order.
     *  scoreartlist: parallel to scorelist, contains art#s of the scores.
     *  scoreartindx: AOFFSETS from art#s are indexes into this array,
     *  	      which contains an index into the score[art]list arrays.
     *  	      Unscored/unsorted articles have a negative # for their
     *  	      entry giving the current scored/sorted status.
     *  scoreflags: One byte (8 bitflags) per article.
     *  This comment needs cleaning up, and should probably get a paragraph
     *  or two in a SA_HACKERS file.
     */
    i = AOFFSET(lastart)+1;	/* number of articles (maximum) */
/* RENAME THE FOLLOWING VARIABLES?!? (use sc_ prefix?) */
    ascorelist = (int*)safemalloc((i+1)*sizeof(int));
    scorelist = ascorelist+1;
    ascoreartlist = (int*)safemalloc((i+1)*sizeof(int));
    scoreartlist = ascoreartlist+1;
    scoreartindx = (int*)safemalloc(i*sizeof(int));
    scoreflags =  (char*)safemalloc(i*sizeof(char));

    /* initialize some of the arrays. */
    for (i=absfirst;i<=lastart;i++) {
/* use AOFFSET directly? */
      int j;
      j = AOFFSET(i);
      scoreartindx[j] = -2;	/* unscored */
      scoreflags[j] = (char)0;	/* all flags unset */
    }
#ifdef HCACHE
    hc_setspin(TRUE);
#endif
    max_indx = -1;
    max_sorted_indx = -1;
    /* let user know if/when there are avail. articles... */
    sc_announce_elig = TRUE;
/* CONSIDER: (for sc_init callers) is lastart properly set yet? */
    sc_fill_min = lastart;
#ifdef SCAN
    if (sa_mode_read_elig)
	sc_fill_read = TRUE;
    else
#endif
	sc_fill_read = FALSE;
/* reindent begin */
    printf("\nScoring articles...") FLUSH;
    fflush(stdout);		/* print it *now* */
    a = lastart;
    /* Note: for proper operation at least one article must be scored */
    /* consider: what if lots of unavailable at end of newsgroup. */
    /* score firstart when only using unread arts? */
    /* use some kind of fast skip ability? (should already occur) */
    while (a>=absfirst) {
	sc_score_art(a,TRUE);	/* I want it *now* */
#ifdef AVAILBITS
	if (is_available(a))
	    break;
	else if (!is_unavailable(a)) {	/* not decided yet */
	    (void)fetchsubj(a,FALSE,FALSE);	/* find out */
	    if (is_available(a))
		break;
	}
#endif
	a--;
    }
    if (a<absfirst) {
	printf("\nNo articles available for scoring\n");
	sc_cleanup();
	return;
    }
    sc_initialized = TRUE;	/* little white lie for lookahead */
    /* now is a good time to load a saved score-list which may exist */
    if (!sc_rescoring)	/* don't load if rescoring */
	sc_load_scores();	/* will be quiet if non-existent */
#ifdef PENDING
    fflush(stdout);
    if (pend_wait) {
	printf("(press key to start reading)");
#else /* !PENDING */
	printf("(interrupt to start reading)");
#endif
	fflush(stdout);
	sc_lookahead(TRUE,TRUE);    /* jump in *now* */
#ifdef PENDING
    }
#endif
    sc_sort();
    printf("\n") FLUSH;
#ifdef HCACHE
    hc_setspin(FALSE);
#endif
    sc_announce_elig = FALSE;
    sc_initialized = TRUE;
}

void
sc_grow(oldlast,newlast)
ART_NUM oldlast,newlast;
{
    int i;

    if (!sc_initialized)
	return;		/* the articles will be scored by sc_init() */
    printf("\nenter: sc_grow(%d,%d)\n",oldlast,newlast) FLUSH;
    fflush(stdout);
    /* make the arrays */
    i = AOFFSET(newlast)+1;	/* number of articles in group */
    ascorelist = (int*)saferealloc((char*)ascorelist,(i+1)*sizeof(int));
    scorelist = ascorelist+1;
    ascoreartlist = (int*)saferealloc((char*)ascoreartlist,(i+1)*sizeof(int));
    scoreartlist = ascoreartlist+1;
    scoreartindx = (int*)saferealloc((char*)scoreartindx,i*sizeof(int));
    scoreflags = (char*)saferealloc((char *)scoreflags,i*sizeof(char));

    /* initialize the grown portion of the arrays */
/* use AOFFSET directly? */
    for (i=oldlast+1;i<=newlast;i++) {
      int j;
      j = AOFFSET(i);
      scoreartindx[j] = -2;	/* unscored */
      scoreflags[j] = (char)0;	/* all flags unset */
    }
    /* remove printouts before release */
    printf("\nGrowing newsgroup: Scoring %d articles...",newlast-oldlast);
    fflush(stdout);
    sc_fill_scorelist(oldlast+1,newlast);
    printf("Done (exit sc_grow)\n");
    fflush(stdout);
}

void
sc_cleanup()
{
    if (!sc_initialized)
	return;		/* we really shouldn't have been called */

    /* possibly auto-save the scores */
    if (sc_savescores) {	/* test for now */
	sc_save_scores();
    }

    sc_loaded_count = 0;	/* 
    /* free the arrays */
    free(ascorelist);
    free(ascoreartlist);
    free(scoreartindx);
    free(scoreflags);

    printf("\nCleaning up scoring...");
    fflush(stdout);
    sf_clean();		/* let the scorefile do whatever cleaning it needs */
    sc_initialized = FALSE;
    printf("Done.\n") FLUSH;
}

void
sc_set_score(a,score)
ART_NUM a;
int score;
{
    if (is_unavailable(a)) {	/* newly unavailable */
	scoreartindx[AOFFSET(a)] = -3;
	return;
    }
    if (kill_thresh_active && (score<=kill_thresh) && (!was_read(a)))
	oneless(a);

    if (scoreartindx[AOFFSET(a)]>-2) {	/* scored before */
	scorelist[scoreartindx[AOFFSET(a)]] = score;	/* update the score */
/* should the scores be resorted? */
	return;
    }
    /* Otherwise, add the new score to the list */
    max_indx++;
    scorelist[max_indx] = score;
    scoreartlist[max_indx] = a;
    scoreartindx[AOFFSET(a)] = -1;	/* not sorted at moment */

    /* let the reader know if we have eligible articles */
    if (sc_announce_elig && ((!was_read(a))
#ifdef SCAN
	|| sa_mode_read_elig
#endif
	)) {
	    printf("!");
	    fflush(stdout);
	    sc_announce_elig = FALSE;
    }
}

#ifdef NEWSFILTER
/* gross, but needed for newsfilter */
static long sc_our_pid = 0;
static char sc_artname[1000];	/* overkill */
#endif

/* CONSIDER: write an "update-score" (already in scorelist) function? */
/* For now use the scorefile routines only.  Hopefully others will
   write more scoring functions. */
/* RENAME? sc_score_basic (?)  What does _art add to the name? */
/* protected against bad input by sc_score_art */
void
sc_score_art_basic(a)
ART_NUM a;
{
    int score;

    score = 0;
#ifdef NEWSFILTER
    if (sc_have_newsclip && sc_use_newsclip) {
#ifdef SERVER
	artopen(a);		/* be sure it is open and created */
	if (sc_our_pid == 0)
	    sc_our_pid = getpid();
	sprintf( sc_artname, "/tmp/rrn%ld.%d", (long)a, sc_our_pid );
	filter_art(sc_artname,(char *)0,ngname,a);
#else /* !SERVER */
	filter_art( spool, ngdir, ngname, a );
#endif /* SERVER */
	score += filter_score_art(a);
    }
#endif /* NEWSFILTER */
    score += sf_score(a);	/* get a score */

    sc_set_score(a,score);	/* set the score */
}

/* CONSIDER: (!) Should the AVAILBITS-using code be #ifdef'ed even though
		 the code will *not* work without it?
 */
/* The routine to call from the outside to score an article.
 * returns 0 if unscored.  Should it do something else special?
 */
/* RENAME? sc_score (?)  What does _art add to the name? */
int
sc_score_art(a,now)
ART_NUM a;
bool_int now;	/* if TRUE, sort the scores if necessary... */
{
    int i;
    if ((a<absfirst) || (a>lastart)) {
	printf("\nsc_score_art: illegal article# %d\n",a) FLUSH;
	sc_keypause();
	return(0);
    }
#ifdef AVAILBITS
    if (is_unavailable(a))
	return(LOWSCORE);
#endif
    i = scoreartindx[AOFFSET(a)];
    switch (i) {
	case -3:	/* unavailable */
	    /* shouldn't be called... */
	    return(LOWSCORE);	/* lowest score */
	case -2:	/* entirely unscored */
	    sc_score_art_basic(a);
#ifdef AVAILBITS
	    if (is_unavailable(a))
		return(LOWSCORE);
#endif
	    /* FALL THROUGH */
	case -1:	/* scored, but not yet sorted */
	    if (!now)
		return(0);	/* no score yet */
	    sc_sort();
	    i = scoreartindx[AOFFSET(a)];
	    /* FALL THROUGH */
	default:	/* scored, sorted article */
	    return(scorelist[i]);
    } /* switch */
}
	
/* CONSIDER: AN OPTION FOR SCORING ONLY UNREAD ARTICLES
	     (obey sc_fill_unread?) */
/* CONSIDER: RETURN # OF ARTICLES SCORED? */
void
sc_fill_scorelist(first,last)
ART_NUM first,last;
{
    int i;

/*  printf("entry: sc_fill_scorelist(%d,%d)\n",first,last) FLUSH; /* */
    for (i=first;i<=last;i++) {
	(void)sc_score_art(i,FALSE);	/* will be sorted later... */
    } /* for */
/*  printf("before sort: sc_fill_scorelist(%d,%d)\n",first,last) FLUSH; /* */
    sc_sort();
/*  printf("entry: sc_fill_scorelist(%d,%d)\n",first,last) FLUSH; /* */
}

/* true if a1 is greater-than a2 *for sorting purposes* */
/* a1 != a2 */
#define SORT_ARTNUMS(a1,a2) \
	((!score_newfirst && (a1>a2)) || (score_newfirst && (a1<a2)))

/* uses heapsort algorithm with inline heap readjustment */
/* score comparison: article a is prior-to article b if either:
 *	score(a)<score(b)  OR:
 *	score(a) == score(b) AND (a<b OR score_newfirst==TRUE)
 * Note: a is *never* == to b.
 */
void
sc_sort()
{
    int i,n;
    int t1,t2;	/* t1 is the scorelist entry, t2 for scoreartlist */
    int j;

/* simply use ascorelist and ascoreartlist directly? */
    sort_scorelist = ascorelist;
    sort_scoreartlist = ascoreartlist;

    n = max_indx+1;
    if (n<1)
	return;		/* nothing to sort */

    for (i=n/2;i>=1;i--) {
	/* begin heap readjust */
	t1 = sort_scorelist[i];
	t2 = sort_scoreartlist[i];

	j = 2*i;
	while(j<=n) {
	    if ((j<n) &&
		((sort_scorelist[j]<sort_scorelist[j+1])
		 || ((sort_scorelist[j] == sort_scorelist[j+1]) &&
		     SORT_ARTNUMS(sort_scoreartlist[j],sort_scoreartlist[j+1])
	       )))
			j++;
	    if ((t1>sort_scorelist[j]) ||
		((t1 == sort_scorelist[j]) &&
		 SORT_ARTNUMS(sort_scoreartlist[j],t2)))
		    break;		/* out of while loop */
	    else {
		sort_scorelist[j/2] = sort_scorelist[j];
		sort_scoreartlist[j/2] = sort_scoreartlist[j];
		j = j*2;
	    }
	} /* while */
	sort_scorelist[j/2] = t1;
	sort_scoreartlist[j/2] = t2;
	/* end heap readjust */
    } /* for */

    for (i=n-1;i>=1;i--) {
	t1 = sort_scorelist[i+1];
	t2 = sort_scoreartlist[i+1];
	sort_scorelist[i+1] = sort_scorelist[1];
	sort_scoreartlist[i+1] = sort_scoreartlist[1];
	sort_scorelist[1] = t1;
	sort_scoreartlist[1] = t2;
	/* begin heap readjust */
	j = 2;
	while(j<=i) {
	    if ((j<i) &&
		((sort_scorelist[j]<sort_scorelist[j+1])
		 || ((sort_scorelist[j] == sort_scorelist[j+1]) &&
		     SORT_ARTNUMS(sort_scoreartlist[j],sort_scoreartlist[j+1])
	       )))
			j++;
	    if ((t1>sort_scorelist[j]) ||
		((t1 == sort_scorelist[j]) &&
		 SORT_ARTNUMS(sort_scoreartlist[j],t2)))
			break;	/* out of while */
	    else {
		sort_scorelist[j/2] = sort_scorelist[j];
		sort_scoreartlist[j/2] = sort_scoreartlist[j];
		j = j*2;
	    }
	} /* while */
	sort_scorelist[j/2] = t1;
	sort_scoreartlist[j/2] = t2;
	/* end heap readjust */
    } /* for */
    /* end of heapsort */

    /* set up scoreartindx */
    for (i=0;i<=max_indx;i++)
	scoreartindx[AOFFSET(scoreartlist[i])] = i;
    max_sorted_indx = max_indx;
}

/* returns the article # with the next higher score, or 0 if no
 * higher one exists
 */
ART_NUM
sc_art_higher_score(a)
ART_NUM a;
{
    int i;

    i = scoreartindx[AOFFSET(a)];
    if (i<0)	/* not scored and/or sorted (shouldn't happen :-) */
	return((ART_NUM)0);
    if (i==max_sorted_indx)
	return((ART_NUM)0);
    else
	return(scoreartlist[i+1]);
}

/* returns the article # with the next lower score, or 0 if no
 * lower one exists
 */
ART_NUM
sc_art_lower_score(a)
ART_NUM a;
{
    int i;

    i = scoreartindx[AOFFSET(a)];
    if (i<0)	/* not scored and/or sorted (shouldn't happen :-) */
	return((ART_NUM)0);
    if (i==0) {
	return((ART_NUM)0);
    }
    else {
	return(scoreartlist[i-1]);
    }
}

/* returns the article# with the highest score (0 if none available) */
ART_NUM
sc_art_highest_score()
{
    if (!sc_initialized)
	return(0);	/* we *really* shouldn't be called! */
    if (max_sorted_indx != max_indx) /* more articles scored */
	sc_sort();
    if (max_sorted_indx<0) { /* no articles sorted */
	printf("ERROR: sc_art_highest_score: no articles sorted.\n") FLUSH;
	return(0);
    }
    return((ART_NUM)scoreartlist[max_sorted_indx]);
}

/* returns the article# with the lowest score (0 if none scored) */
ART_NUM
sc_art_lowest_score()
{
    if (!sc_initialized)
	return(0);	/* we *really* shouldn't be called! */
    if (max_sorted_indx != max_indx)	/* more articles scored */
	sc_sort();
    if (max_sorted_indx<0) { /* no articles sorted */
	printf("ERROR: sc_art_lowest_score: no articles sorted.\n") FLUSH;
	return(0);
    }
    return((ART_NUM)scoreartlist[0]);
}

/* consider having this return a flag (is finished/is not finished) */
void
sc_lookahead(flag, nowait)
/* flag not used now, but when used... */
bool_int flag;		/* TRUE means sort now
			 * FALSE means wait until later... */
bool_int nowait;	/* TRUE means start scoring immediately (group entry)
			 * FALSE means use NICEBG if available */
{
    ART_NUM oldart = openart;
    ART_POS oldartpos;

#ifdef PENDING
    if (input_pending())
	return;		/* delay as little as possible */
#endif
    if (!sc_initialized)
	return;		/* don't score then... */
    if (max_indx != max_sorted_indx)
	sc_sort();	/* sort the scores if not sorted */
#ifdef PENDING
#ifdef NICEBG
    if (sc_mode_nicebg && !nowait)
	if (wait_key_pause(10))		/* wait up to 1 second for a key */
	    return;
#endif
#endif
    if (oldart)			/* Was there an article open? */
	oldartpos = ftell(artfp);	/* where were we in it? */
#ifndef PENDING
    if (int_count)
	int_count = 0;		/* clear the interrupt count */
#endif
    while (
#ifdef PENDING
	   !input_pending() &&
#endif
	   (sc_fill_min>absfirst)) {
#ifndef PENDING
	if (int_count>0) {
	    int_count = 0;
	    return;	/* user requested break */
	}
#endif
	/* shall we leave early? */
	if ((!sc_fill_read) && (sc_fill_min<=firstart)) {
	    sc_sort();
	    break;	/* out of while loop */
	}
	if ((max_indx-max_sorted_indx)>scorehunksize) {
	    sc_sort();
	    continue;	/* with while loop */
	}
	sc_fill_min--;
	if (!sc_fill_read)	/* score only unread */
	    while (was_read(sc_fill_min) &&
	      (sc_fill_min>firstart))
		sc_fill_min--;
	(void)sc_score_art(sc_fill_min,FALSE);
    }
    if (sc_fill_min==absfirst)
	sc_sort();		/* sort scores if reached end. */
    if (oldart) {	    /* copied from cheat.c */
	artopen(oldart);
	fseek(artfp,oldartpos,0);	/* do not screw the pager */
    }
}

/* move this elsewhere? util.c perhaps? give it a string argument to print? */
void
sc_keypause()
{
    printf("\nSCORE: Press interrupt key to continue\n") FLUSH;
    pause();
    int_count = 0;
}

int
sc_percent_scored()
{
    int i,total,scored;

    if (!sc_initialized)
	return(0);	/* none scored */
    if (sc_fill_min==absfirst)
	return(100);
    if ((sc_fill_min<=firstart)
#ifdef SCAN
	&& !sa_mode_read_elig
#endif
       )
	return(100);
    i = firstart;
#ifdef SCAN
    if (sa_mode_read_elig)
	i = absfirst;
#endif
    total = scored = 0;
    for (;i<=lastart;i++) {
	if (is_unavailable(i))
	    continue;
	if (was_read(i)
#ifdef SCAN
	    && !sa_mode_read_elig
#endif
	   )
		continue;
	total++;
	if (scoreartindx[AOFFSET(i)]>-1)	/* not scored and sorted */
	    scored++;
    } /* for */
    if (total == 0)
	return(0);
    return((scored*100)/total);
}

void
sc_rescore_arts()
{
    ART_NUM a;
    for (a=lastart;a>=absfirst;a--)
	if (scoreartindx[AOFFSET(a)]>-2)	/* scored article */
	    sc_score_art_basic(a);		/* rescore it then */
    sc_sort();
}

/* Wrapper to isolate scorefile functions from the rest of the world */
/* corrupted (:-) 11/12/92 by CAA for online rescoring */
void
sc_append(line)
char *line;
{
    char filechar;
    if (!line || !*line)		/* empty line */
	return;

    filechar = *line;			/* first char */
    sf_append(line);
    if (filechar == '!') {
	printf("\nRescoring articles...");
	fflush(stdout);
	sc_rescore_arts();
	printf("Done.\n") FLUSH;
#ifdef SCAN
	if (sa_initialized)
	    *s_top_ent = -1;		/* reset top of page */
#endif
    }
}

/* load/save scores. */
void
sc_load_scores()
{
    ART_NUM a;
    char *s;
    int score;
    char lbuf[256];	/* for reading in scores */
    FILE *fp;
    int loaded,used;
    char *gname;

    loaded = used = 0;
    sc_loaded_count = 0;
    s = getval("SAVESCORES","///");
    if (sc_savefile) {
	switch(*sc_savefile) {
	    case '/':
	    case '~':
	    case '%':
		s = sc_savefile;
		break;
	    case ':':		/* from SAVESCORES directory */
		if (strEQ(s,"///"))
		    return;		/* no value here */
		strcpy(lbuf,s);
		s = rindex(lbuf,'/');
		if (!s) {		/* error */
		    printf("\nSAVESCORES has no filename component.\n") FLUSH;
		    return;
		}
		s++;
		strcpy(s,sc_savefile+1);
		s = lbuf;
		break;
	    default:
		printf("\nSAVESCORES: bad filename %s.\n",sc_savefile) FLUSH;
		return;
	}
    }
    if (strEQ(s,"///"))
	return;			/* no valid score file name */
    fp = fopen(filexp(s),"r");
    if (!fp) {
/*	printf("Could not open score save file for reading.\n") FLUSH; /* */
	return;
    }
    gname = savestr(filexp("%C"));
    if (sc_savefile) {
	/* merged type, so skip until filename */
	while (fgets(lbuf,250,fp))
	    if (*lbuf=='!') {
		lbuf[strlen(lbuf)-1] = '\0';	/* strip \n */
		if (strEQ(lbuf+1,gname))
		    break;
	    }
	if (!lbuf || !*lbuf || !strEQ(lbuf+1,gname)) {
	    free(gname);
	    return;
	}
    }
    printf("\nLoading scores...");
    fflush(stdout);
/* savescore format:
 * !news.group.name	"!comp.sys.amiga.misc"
 * :article_num		":123456"
 * score		"21" "-17"
 *
 * Consider some error checking later.  For instance, what if someone
 * saved alt.sex without setting the filename and set the save filename
 * for alt.sex.bondage to alt.sex.  (Loading alt.sex could fail then.)
 * Perhaps give an error if no filename and a '!' is found?
 */
    while (fgets(lbuf,250,fp)) {
	if (*lbuf==':') {
	   a = atoi(lbuf+1);	/* set the article # */
	   continue;
	}
	if (*lbuf=='!') {	/* group of shared file */
	    /* strip the \n */
	    lbuf[strlen(lbuf)-1] = '\0';
	    if (strEQ(lbuf+1,gname))	/* this group */
		continue;
	    break;
	}
	score = atoi(lbuf);
	loaded++;
	if ((a<=lastart) && (a>=absfirst) && (!is_unavailable(a)) &&
	    (!was_read(a)) && scoreartindx[AOFFSET(a)] == -2) {
		sc_set_score(a,score);
	    used++;
	}
	a++;
    }
    fclose(fp);
    free(gname);
    sc_sort();
    printf("Done.  ");
    /* sloppy plurals (:-) */
    printf("(%d scores loaded, %d scores used)\n",loaded,used) FLUSH;
    sc_loaded_count = loaded;
#ifdef SCAN
    if (sa_initialized)
	*s_top_ent = -1;	/* reset top of page */
#endif
}

void
sc_save_scores()
{
    ART_NUM a,last;
    char *s;
    FILE *tmpfp;
    FILE *savefp;		/* for reading the destination */
    int saved;
    bool skipping;		/* skipping the old scores? */
    bool found_group;		/* found at least one group name */
    char *gname,*tname;
    char *savename;
    char lbuf[512];
    char lbuf2[512];		/* what's another buffer between... */
    bool w_error;		/* write error occurred */

    gname = savestr(filexp("%C"));
    saved = 0;
    last = 0;
    skipping = FALSE;
    found_group = FALSE;
    w_error = FALSE;

    s = getval("SAVESCORES","///");
    if (sc_savefile) {
	switch(*sc_savefile) {
	    case '/':
	    case '~':
	    case '%':
		s = sc_savefile;
		break;
	    case ':':		/* from SAVESCORES directory */
		if (strEQ(s,"///"))
		    return;		/* no value here */
		strcpy(lbuf,s);
		s = rindex(lbuf,'/');
		if (!s) {		/* error */
		    printf("\nSAVESCORES has no filename component.\n") FLUSH;
		    return;
		}
		s++;
		strcpy(s,sc_savefile+1);
		s = lbuf;
		break;
	    default:
		printf("\nSAVESCORES: bad filename %s.\n",sc_savefile) FLUSH;
		free(gname);
		return;
	}
    }
    if (strEQ(s,"///")) {
	printf("error: SAVESCORES undefined and no explicit filename.\n") FLUSH;
	free(gname);
	return;
    }
    strcpy(lbuf2,filexp(s));
    tname = rindex(lbuf2,'/');
    if (!tname) {
	printf("\nSAVESCORES has no filename component.\n") FLUSH;
	return;
    }
    tname++;
    strcpy(tname,filexp("tmpscores.%$"));
    tname = lbuf2;
    tmpfp = fopen(tname,"w");
    if (!tmpfp) {
	printf("Could not open score save file %s for writing.\n",tname) FLUSH;
	free(gname);
	return;
    }
    savename = savestr(filexp(s));
    savefp = fopen(savename,"r");
    /* not being able to open is OK */
    if (savefp) {
	/* copy the old file, except for the current section */
	while (fgets(lbuf,250,savefp)) {
	    if (*lbuf=='!') {	/* group of shared file */
		/* strip the \n */
		lbuf[strlen(lbuf)-1] = '\0';
		found_group = TRUE;
		if (strEQ(lbuf+1,gname)) {	/* this group */
		    skipping = TRUE;
		    continue;
		}
		/* ...and put it back */
		lbuf[strlen(lbuf)] = '\n';
		skipping = FALSE;
	    }
	    if (!skipping) {
		fprintf(tmpfp,"%s",lbuf);	/* includes \n */
		if (ferror(tmpfp)) {
		    w_error = TRUE;
		    goto write_error;
		}
	    }
	}
	fclose(savefp);
    }
    if (!found_group) {
	/* this must be an old style which doesn't have group names. */
	/* rewrite the scores to add group names */
	fclose(tmpfp);
	tmpfp = fopen(tname,"w");
	if (!tmpfp) {
	    printf("Could not open score save file %s for writing.\n",
		   tname) FLUSH;
	    free(gname);
	    free(savename);
	    return;
	}
    }
    sc_sort();	/* make sure all scored are sorted */
    fprintf(tmpfp,"!%s\n",gname);	/* add the header */
    for (a=firstart;a<=lastart;a++)
	if ((!was_read(a)) && (scoreartindx[AOFFSET(a)]>=0)) {
	    if ((last==0) || (last!=a-1))    /* print article # */
		fprintf(tmpfp,":%ld\n",a);
	    /* print article's score */
	    fprintf(tmpfp,"%ld\n",scorelist[scoreartindx[AOFFSET(a)]]);
	    last = a;
	    saved++;
	    if (ferror(tmpfp)) {
		w_error = TRUE;
		goto write_error;
	    }
	}
    fflush(tmpfp);
    fclose(tmpfp);
    makedir(savename,MD_FILE);	/* make any necessary directories */
/* CAA 1/7/93: I'm really not sure about the code below.  When I put an
 * UNLINK here (at this comment's location), I get .nfs* files frequently.
 * Since my manual for rename(2) specifies that the destination is
 * overwritten, the code here simply renames if the rename() system
 * is available.  (The link(2) page is very unclear about if replacement
 * is valid, and seems to hint against, so I put the UNLINK call there.
 */
#ifdef RENAME
    rename(tname,savename);
#else
    UNLINK(savename);
    link(tname,savename);
    UNLINK(tname);
#endif
write_error:
    if (w_error) {		/* error in writing the tmp file */
	printf("Write error in temporary saved score file.\n") FLUSH;
	printf("(Old saved scores kept)\n") FLUSH;
	fclose(tmpfp);	/* won't have been done on an error */
	UNLINK(tname);
    }
    free(gname);
    free(savename);
    if (w_error)
	return;
    /* sloppy plurals (:-) */
    printf("%d scores saved.\n",saved) FLUSH;
}

void
sc_cmd_ext(line)
char *line;
{
    int i;
    char *s;

    switch (*line) {
	case 's':	/* save scores */
	    sc_save_scores();
	    break;
	case 'l':	/* load scores */
	    sc_load_scores();
	    break;
	default:
	    printf("Unknown scoring command |%s|\n",line);
    }
}

void
sc_rescore()
{
    sc_rescoring = TRUE;	/* in case routines need to know */
    sc_cleanup();	/* get rid of the old */
    sc_init(TRUE);	/* enter the new... (wait for rescore) */
#ifdef SCAN
    if (sa_initialized)
	*s_top_ent = -1;	/* reset top of page */
#endif
    sc_rescoring = FALSE;
}

/* May have a very different interface in the user versions */
void
sc_score_cmd(line)
char *line;
{
    int i;
    char *s;

    switch (*line) {
        case 'f':	/* fill (useful when PENDING is unavailable) */
	    printf("Scoring more articles...");
	    fflush(stdout);	/* print it now */
#ifdef HCACHE
	    hc_setspin(TRUE);
#endif
	    sc_lookahead(TRUE,FALSE);
#ifdef HCACHE
	    hc_setspin(FALSE);
#endif
	    /* consider a "done" message later,
	     * *if* lookahead did all the arts */
	    printf("\n") FLUSH;
	    break;
	case 'r':	/* rescore */
	    printf("Rescoring articles...\n") FLUSH;
	    sc_rescore();
	    break;
	case 's':	/* verbose score for this article */
/* CONSIDER: A VERBOSE-SCORE ROUTINE (instead of this hack) */
	    sf_score_verbose = TRUE;
	    i = sf_score(art);
	    sf_score_verbose = FALSE;
	    printf("Total score is %d\n",i) FLUSH;
	    break;
	case 'd':	/* dump */
	    sf_dump(line+1);
	    sc_keypause();
	    break;
#ifdef NEWSFILTER
	case 'n':	/* talk to news filter (Newsclip) */
	    for (s=line+1;((*s==' ')||(*s=='\t'));s++)
		; /* EMPTY */
	    /* make printing optional?  or just remove? */
	    if (filter_command(s))
		printf("Filter command accepted.\n");
	    else
		printf("Filter command rejected.\n");
	    break;
#endif
	case 'e':	/* edit scorefile or other file */
	    for (s=line+1;((*s==' ')||(*s=='\t'));s++)
		; /* EMPTY */
	    if (!*s)	/* empty name for scorefile */
		sf_edit_file("\"");	/* edit local scorefile */
	    else
		sf_edit_file(s);
	    break;

#ifdef UNDEF
	case 't':	/* test vector */
	    /* may do something later */
#endif
	case 'x':	/* extension */
	    sc_cmd_ext(line+1);
	    break;
	default:
	    printf("Unknown scoring command |%s|\n",line);
    } /* switch */
}

void
sc_kill_threshold(thresh)
int thresh;		/* kill all articles with this score or lower */
{
    int i;

    for (i=0;i<=max_sorted_indx;i++) {
	if ((scorelist[i]<=thresh) &&
#ifdef SCAN
		(sa_basic_elig(scoreartlist[i])) &&
#endif
		(!was_read(scoreartlist[i]))) {

			oneless(scoreartlist[i]);
	} /* if */
    } /* for */
}

/* print the scoreflags too? */
void
sc_dumpscores()
{
    int j;

    printf("max_indx: %d\n",max_indx) FLUSH;
    printf("max_sorted_indx: %d\n",max_sorted_indx) FLUSH;
    /* Note: Column header widths do not match data widths.
	     Just another ugly debugging tool. :-) */
    printf("| sl | sal | sai | sai[sal] | j\n");
    for (j=0;j<=max_indx;j++) {
	printf("| %7d | %7d | %4d | %4d | %4d\n",
	scorelist[j],scoreartlist[j],scoreartindx[j],
	scoreartindx[AOFFSET(scoreartlist[j])],j) FLUSH;
    }
}
#endif /* SCORE */
