/* sgcmd.c
 *
 * ScanGroup command interpreter
 */

#include "EXTERN.h"
#include "common.h"
#include "final.h"		/* assert() */
#include "scan.h"
#include "scmd.h"
#include "smisc.h"
#include "spage.h"
#include "sgroup.h"
#include "sgdata.h"
#include "sgfile.h"
#include "sgmisc.h"
#include "term.h"
#include "INTERN.h"
#include "sgcmd.h"

static char sg_fakebuf[256];

void
sg_go_bot()
{
    *s_ref_all = TRUE;			/* help uses whole screen */
    s_goxy(0,LINES-*s_bot_lines);	/* go to bottom bar */
    erase_eol();			/* erase to end of line */
    s_goxy(0,LINES-*s_bot_lines);	/* go (back?) to bottom bar */
    fflush(stdout);			/* do it *now* */
}

/* interprets command in TRN-global buf */
int
sg_docmd()
{
    long a;		/* entry pointed to */
    long b;		/* misc. artnum */
    int i,j;		/* for misc. purposes */
    bool flag;		/* misc */

    a = page_ents[*s_ptr_page_line].entnum;
    switch(*buf) {
	case 'Q':	/* quit from ScanGroup to old newsgroup level */
	    sg_loop = FALSE;
	    sg_go_bot();
	    return(SG_RETURN);
	case '+':	/* read group by command */
	case ';':
	case '\n':
	case ' ':	/* enter group/context by this command */
	    sg_go_bot();
	    switch (sg_ents[a].type) {
		case 1:		/* normal newsgroup */
		    if ((*buf == ' ') || (*buf == '\n')) {
			/* the default if extended modes unavailable */
			*buf = '=';
#ifdef SCAN
			*buf = ';';
#else
#ifdef USETHREADS
			*buf = '+';
#endif
#endif
		    }
		    sg_fakebuf[0] = *buf;
		    sg_fakebuf[1] = '\0';
		    sg_fakecmd = sg_fakebuf;
		    return(a);
		    break;
		case 2:		/* file */
		    (void)sg_use_file(sg_ents[a].restrict,
				sg_ents[a].name);
		    if (s_fillpage()<=0) {
			/* if we cannot fill a page, go back up */
			s_beep();		/* alert the user */
			sg_old_context();	/* go back up */
			*s_refill = TRUE;
			*s_ref_all = TRUE;	/* refill and refresh all */
		    }
		    return(0);	/* carry on... */
		case 3:		/* restriction list */
		    sg_use_restrict(sg_ents[a].restrict,
				     sg_ents[a].name);
		    if (s_fillpage()<=0) {
			/* if we cannot fill a page, go back up */
			s_beep();		/* alert the user */
			sg_old_context();	/* go back up */
			*s_refill = TRUE;
			*s_ref_all = TRUE;	/* refill and refresh all */
		    }
		    return(0);	/* carry on... */
	    }
	    printf("Entry pointed to has an illegal type!\n") FLUSH;
	    assert(FALSE);
	    /* NOTREACHED */
	case 'q':	/* quit from context or strn */
	    /* maybe make behavior optional (sg_fastquit variable?) */
	    if (sg_num_contexts==1) {		/* at top */
		sg_loop = FALSE;
		sg_fakecmd = "q";
		sg_go_bot();
		return(SG_FAKE);
	    }
	    /* FALL THROUGH */
	case '(':	/* go up a context */
	    if (sg_num_contexts<2) {	/* already at top */
		s_beep();
		break;
	    }
	    sg_go_bot();
	    sg_old_context();	/* maybe call a wrapper later */
	    *s_refill = TRUE;
	    *s_ref_all = TRUE;	/* refill and refresh all */
	    break;
	case Ctl('e'):		/* edit current file */
	    for (i=sg_num_contexts-1;i>=0;i--)
		if (sg_contexts[i].filename)	/* EQUALS */
		    break;
	    if (i<0) {
		s_beep();
		break;
	    }
	    sg_edit_file();
	    *s_ref_all = TRUE;
	    break;
	case 'L':	/* display newsgroup descriptions */
	    if (sg_desc_unavail) {
		sg_go_bot();
		printf("Newsgroup descriptions are unavailable now.\n") FLUSH;
		printf("Press any key to continue\n");
		getcmd(buf);
		eat_typeahead();
		break;
	    }
	    sg_use_desc = !sg_use_desc;
	    if (sg_use_desc)
		sg_look_num = 0;	/* reset lookahead */
	    *s_refill = TRUE;
	    *s_ref_all = TRUE;	/* refill and refresh all */
	    break;
	default:
	    s_beep();
	    break;
    }
    return(0);
}
