/* sgdata.h
 *
 * Data structures and functions for ScanGroups
 */

/* per-newsgroup data */
typedef struct sg_groupdata {
    char *name;
    char *desc;		/* description of newsgroup */
    long rcnum;		/* index into rc array */
    long flags;
    /* current flag values:  (flags default to 0/FALSE)
     *     1	moderated newsgroup
     *     2	newsgroup's toread[] has been set
     *     4	Attempted to fill description field
     */
} SG_GROUPDATA;

EXT int sg_num_groups		INIT(0);
EXT int sg_groups_alloc		INIT(0);
EXT SG_GROUPDATA *sg_groups	INIT((SG_GROUPDATA*)NULL);

/* if true, have tried to get the newsgroup descriptions and failed */
EXT bool sg_desc_unavail	INIT(FALSE);

/* per-entry data */
/* could be a union, but why bother? */
typedef struct sg_entrydata {
    int  type;
    /* 1: normal newsgroup
     * 2: file
     * 3: restriction list
     */
    long flags;
    /* for initial testing, groupnum is newsgroup number */
    long groupnum;		/* index into sg_groups */
    char *name;			/* what to display for the description */
    /* find a better name for the below */
    char *restrict;		/* filename or restriction list */
} SG_ENTRYDATA;

/* note: the 0th entry of sg_ents is unused (scan compatibility) */
EXT int sg_num_ents		INIT(0);
EXT int sg_ents_alloc		INIT(0);
EXT SG_ENTRYDATA *sg_ents	INIT((SG_ENTRYDATA*)NULL);

typedef struct sg_context {
    int  cnum;			/* context number */
    int  num_ents;		/* number of entries */
    int  ents_alloc;		/* number of allocated entries */
    char *title;		/* for this context */
    char *filename;		/* if non-null, is filename of local context */
    /* add other things like flags later */
    SG_ENTRYDATA *ents;		/* entries */
} SG_CONTEXT;

EXT int sg_num_contexts		INIT(0);
EXT int sg_max_contexts		INIT(0);
EXT SG_CONTEXT *sg_contexts	INIT((SG_CONTEXT*)NULL);

void	sg_init_context	ANSI((void));
void	sg_new_context	ANSI((void));
void	sg_old_context	ANSI((void));

bool	sg_init_ents	ANSI((void));
int	sg_add_ent	ANSI((int,char*));
void	sg_free_ents	ANSI((void));
void	sg_delete_invalid_ents	ANSI((void));

void	sg_init_groups	ANSI((void));
void	sg_reinit_groups	ANSI((void));
bool	sg_add_ng	ANSI((int));
bool	sg_go_ng	ANSI((char *));
void	sg_toread	ANSI((long,bool_int));
void	sg_allgroups	ANSI((void));
bool	sg_group_desc	ANSI((int));
