/* sgmisc.c
 *
 * ScanGroup misc. functions
 * (includes many of the Scan interface routines)
 */

/* trim this list */
#include "EXTERN.h"
#include "common.h"
#include "ng.h"		/* for NG_PREV, etc... */
#include "rn.h"
#include "rcln.h"
#include "rcstuff.h"
#include "scan.h"
#include "scmd.h"
#include "sdisp.h"
#include "sgroup.h"
#include "sgdata.h"
#include "term.h"
#include "util.h"
#include "INTERN.h"
#include "sgmisc.h"

bool
sg_eligible(ent)
long ent;
{
    long num;
    switch (sg_ents[ent].type) {
	case 2:		/* file type */
	    return(TRUE);
	case 3:		/* restriction list */
	    return(TRUE);
    }
    /* check newsgroup */
    sg_toread(sg_ents[ent].groupnum,FALSE);
    num = toread[sg_groups[sg_ents[ent].groupnum].rcnum];
    if (num>0)	/* unread articles here */
	return(TRUE);
    return(FALSE);
}

long
sg_first()
{
    if (sg_num_ents)
	return(1);
    return((long)0);
}

long
sg_last()
{
    if (sg_num_ents)
	return((long)sg_num_ents-1);
    return((long)0);
}

long
sg_next(ent)
long ent;
{
    if (ent==(sg_num_ents-1))
	return((long)0);
    return(ent+1);
}

long
sg_prev(ent)
long ent;
{
    if (ent<=1)
	return((long)0);
    return(ent-1);
}

char *
sg_get_desc(ent,line)
long ent;
int line;
{
    static char desc_buf[256];
    static char lbuf[256];
    char *s;
    int num;
    int grp;

    desc_buf[0] = '\0';
    switch (line) {
	case 1:
	    /* name has same function for any type */
	    sprintf(lbuf,"%s",sg_ents[ent].name);
	    strcat(desc_buf,lbuf);
	    break;
	case 2:
	    s = sg_groups[sg_ents[ent].groupnum].desc;
	    if (!s || !*s)
		strcpy(desc_buf,"<No group description>");
	    else
		strcpy(desc_buf,s);
	    break;
    }
    desc_buf[*s_desc_cols] = '\0';
    return(desc_buf);
}

/* fixed 10-character line */
char *
sg_get_statchars(ent,line)
long ent;
int line;
{
    static char statbuf[256];
    int num;
    int grp;
    int i;
    char *p;

    if (line==2) {
	strcpy(statbuf,"          ");
	return(statbuf);
    }
    switch (sg_ents[ent].type) {
	case 2:		/* file */
	    /* strings copied so they can be altered */
	    strcpy(statbuf,"      file");
	    return(statbuf);
	case 3:		/* restriction list */
	    strcpy(statbuf,"      list");
	    return(statbuf);
    }
    grp = sg_groups[sg_ents[ent].groupnum].rcnum;
    for (i=0;i<5;i++)
	statbuf[i] = '.';
    p = statbuf;
    sg_toread(sg_ents[ent].groupnum,FALSE);
    switch (toread[grp]) {
	case TR_JUNK:
	    *p = 'J';
	    break;
	case TR_BOGUS:
	    *p = 'B';
	    break;
	case TR_UNSUB:
	    *p = '!';
	    break;
	case TR_NONE:
	    *p = '-';
	    break;
	default:
	    *p = '+';
	    break;
    }
    if (sg_groups[sg_ents[ent].groupnum].flags & 1)	/* moderated */
	statbuf[1] = 'M';
    num = toread[grp];
    if (num>TR_UNSUB) {
	if (num<100000) {	/* will fit in 5 columns */
	     sprintf(statbuf+5,"%5ld",(long)num);
	} else {
	     sprintf(statbuf+5,"%s","+++++");
	}
    } else {
	/* unsub, bogus, and junk don't get counts */
	sprintf(statbuf+5,"%s","*****");
    }
    statbuf[*s_status_cols] = '\0';
    return(statbuf);
}

void
sg_refresh_top()
{
    char lbuf[256];
    char *s;

    standout();
    if (sg_num_contexts==1)
	printf("Top Level |");
    else
	printf("Level %d |",sg_num_contexts);
    /* print up to 50 characters of title (change to less later?) */
    s = sg_contexts[sg_num_contexts-1].title;
    if (strlen(s)<51)
	printf(" %s",s);
    else {
	/* actually use 53 characters of line */
	strncpy(lbuf,s,50);
	lbuf[50] = '\0';
	printf(" %s...",lbuf);
    }
    un_standout();
    erase_eol();
    printf("\n") FLUSH;
}

void
sg_refresh_bot()
{
    standout();
    s_mail_and_place();
    un_standout();
    erase_eol();
    fflush(stdout);
}

int
sg_ent_lines(ent)
long ent;
{
    if (sg_use_desc) {
	switch (sg_ents[ent].type) {
	    case 1:		/* newsgroup */
		if (sg_group_desc(sg_ents[ent].groupnum))
		    return(2);
		/* FALL THROUGH */
	    case 2:		/* filename */
	    case 3:		/* list */
		return(1);
	}
    }
    return(1);
}

/* returns TRUE if screen/window is big enough, FALSE otherwise */
/* for now, the requirements are 15 columns and 5 lines */
bool
sg_check_screen()
{
    if (LINES<5 || COLS<5)
	return(FALSE);
    return(TRUE);
}

/* sets up shape of screen */
/* assumes s_initscreen has been called */
void
sg_set_screen()
{
    /* One size fits all for now. */
    *s_top_lines = 1;
    *s_bot_lines = 1;
    *s_status_cols = 10;
    *s_cursor_cols = 2;
    /* (scr_width-1) keeps last character blank. */
    *s_desc_cols = (scr_width-1)-*s_status_cols-*s_cursor_cols;
}

/* do something useful related to the scangroup lookahead */
void
sg_lookahead()
{
#ifdef PENDING
    long flags;

    if (wait_key_pause(10))	/* spend one second waiting for a key */
	return;			/* give user commands first priority */
    while (sg_look_num<sg_num_groups) {
	if (input_pending())
	    return;
	flags = sg_groups[sg_look_num].flags;
	if (!(flags & 2))
	    sg_toread(sg_look_num,FALSE);
	if (sg_use_desc && !(flags & 4))
	    (void)sg_group_desc(sg_look_num);
	sg_look_num++;
    }
#endif
}
