/* sgroup.c
 *
 * Scan Newsgroup main
 */

#include "EXTERN.h"
#include "common.h"
#include "ng.h"		/* for NG_PREV, etc... */
#include "rn.h"
#include "rcln.h"
#include "rcstuff.h"
#include "scan.h"
#include "scmd.h"
#include "sdisp.h"
#include "smisc.h"
#include "spage.h"
#include "sgdata.h"
#include "sgfile.h"
#include "sgmisc.h"
#include "term.h"
#include "util.h"
#include "INTERN.h"
#include "sgroup.h"

int
sg_main()
{
    if (sg_initialized)
	return(sg_continue());
    if (!sg_init())		/* failed to initialize */
	return(SG_RETURN);
    return(sg_continue());
}

/* returns TRUE for successful init, false otherwise */
bool
sg_init()
{
    char lbuf[256];

    /* check to see if screen is big enough */
    if (!sg_check_screen()) {
	printf("Sorry--screen too small.\n") FLUSH;
	return(FALSE);
    }
    sg_init_context();

    /* read in all newsgroups in newsrc */
    sg_init_groups();
    /* early initialization of ents */
/*    sg_init_ents(); */

    /* first try customized, then all groups */
    strcpy(lbuf,getval("SCANGROUPS","%p/scangroups"));
    strcat(lbuf,"/top");
    if (!sg_use_file(lbuf,"Top Level"))
	sg_allgroups();
/* PROBLEM: this initialization/check is quite late... */
    if (s_initscreen()) {		/* If not able to init screen...*/
/* de-initialize later */
	return(FALSE);			/* ...most likely dumb terminal */
    }
    sg_initialized = TRUE;
    return(TRUE);
}


int
sg_continue()
{
    char sg_oldmode;
    int i;
    long e;

    if (sg_changed_newsrc)
      sg_reinit_groups();
    s_change_context(sg_contexts[sg_num_contexts-1].cnum);
    /* see if we should move on to next group */

    e = 0;	/* also a flag */
    switch (sg_ngstatus) {
        /* case codes modified from scmd.c */
	case NG_SELPRIOR:
	    if (*s_ptr_page_line>0) { /* more on page... */
		*s_ptr_page_line -=1;
		e = page_ents[*s_ptr_page_line].entnum;
	    }
	    else {
		if (!s_prev_elig(page_ents[0].entnum)) {
		    e = 0;
		} else {
		    s_go_prev_page();
		    *s_ptr_page_line = *s_bot_ent; /* go to page bot. */
		    e = page_ents[*s_ptr_page_line].entnum;
		}
	    }
	    break;
	case NG_SELNEXT:
	    if (*s_ptr_page_line<(*s_bot_ent)) { /* more on page... */
		*s_ptr_page_line +=1;
		e = page_ents[*s_ptr_page_line].entnum;
	    }
	    else {
		if (!s_next_elig(page_ents[*s_bot_ent].entnum)) {
		    e = 0;
		} else {
		    s_go_next_page();	/* will jump to top too... */
		    e = page_ents[*s_ptr_page_line].entnum;
		}
	    }
	    break;
    }
    if (e && (sg_ents[e].type==1)) {	/* going to another newsgroup entry */
	(void)sg_go_ng(sg_groups[sg_ents[e].groupnum].name);
	sg_lastent_read = e;
	printf("Going to newsgroup %s\n",
		sg_groups[sg_ents[e].groupnum].name) FLUSH;
	/* nice trick: a fake command will be set up already (?) */
	return(SG_FAKE);
    }	
    /* redraw it *all* */
    *s_ref_all = TRUE;
    if ((*s_top_ent<1) || (*s_top_ent>=sg_num_ents))
	*s_top_ent = s_first();
    i = s_fillpage();
    if ((i==-1) || (i == 0)) {
	printf("s_fillpage could not get a page\n") FLUSH;
	if (sg_num_contexts>1) {
	    sg_old_context();
	    *s_refill = TRUE;
	    return(sg_continue());
	}
	/* for now just leave if no page could be filled. */
	sg_loop = 0;	/* go back */
	return(SG_RETURN);
    }
    sg_oldmode = mode;			/* save mode */
    mode = 'g';				/* for RN macros */
    i = s_cmdloop();
    mode = sg_oldmode;			/* restore mode */
    if (i>0) {		/* an entry is selected */
	/* for moment, all names are same as group names */
	sg_go_ng(sg_groups[sg_ents[i].groupnum].name);
	sg_lastent_read = i;
	printf("Going to newsgroup %s\n",
		sg_groups[sg_ents[i].groupnum].name) FLUSH;
	/* the command will set the fake */
	return(SG_FAKE);
    }
    /* something else (quit, return, etc...) */
    if (i==SG_QUIT) {
	sg_loop = 0;
	/* later deinitialize */
	return(SG_RETURN);
    }
    return(i);
}
